from __future__ import annotations

from collections.abc import Callable
from statistics import mean

from .hierarchy import Node


def default_separation(a: Node, b: Node) -> int:
    return 1 if a.parent == b.parent else 2


def mean_x(children: list[Node]) -> float:
    return mean((child.x for child in children))


def max_y(children: list[Node]) -> float:
    return 1 + max((child.y for child in children))


def leaf_left(node: Node) -> Node:
    while children := node.children:
        node = children[0]
    return node


def leaf_right(node: Node) -> Node:
    while children := node.children:
        node = children[-1]
    return node


class Cluster:
    """
    Cluster layout
    """

    def __init__(self):
        self._separation = default_separation
        self._dx = 1
        self._dy = 1
        self._node_size = False

    def __call__(self, root: Node) -> Node:
        """
        Lays out the specified root hierarchy, assigning the following
        properties on root and its descendants:

        * :code:`node.x` - the x-coordinate of the node
        * :code:`node.y` - the y coordinate of the node

        The coordinates :code:`x` and :code:`y` represent an arbitrary
        coordinate system; for example, you can treat :code:`x` as an angle and
        :code:`y` as a radius to produce a radial layout. You may want to call
        :code:`root.sort` before passing the hierarchy to the cluster layout.

        Parameters
        ----------
        root : Node
            Root node

        Returns
        -------
        Node
            Node organized as cluster
        """
        x = 0
        previous_node = None

        def cluster(node: Node):
            nonlocal previous_node, x
            children = node.children
            if children:
                node.x = mean_x(children)
                node.y = max_y(children)
            else:
                if previous_node is None:
                    node.x = 0
                else:
                    x += self._separation(node, previous_node)
                    node.x = x
                node.y = 0
                previous_node = node

        root.each_after(cluster)

        left = leaf_left(root)
        right = leaf_right(root)
        x0 = left.x - self._separation(left, right) * 0.5
        x1 = right.x + self._separation(right, left) * 0.5

        if self._node_size:

            def update(node):
                node.x = (node.x - root.x) * self._dx
                node.y = (root.y - node.y) * self._dy
        else:

            def update(node):
                node.x = (node.x - x0) / (x1 - x0) * self._dx
                node.y = (1 - ((node.y / root.y) if root.y else 1)) * self._dy

        return root.each_after(update)

    def set_separation(self, separation: Callable[[Node, Node], int]) -> Cluster:
        """
        Sets the :code:`separation` accessor to the specified function and
        returns this cluster layout.

        Parameters
        ----------
        separation : Callable[[Node, Node], int]
            Separation function which compares two nodes.

        Returns
        -------
        Cluster
            Itself
        """
        self._separation = separation
        return self

    def set_size(self, size: tuple[float, float]) -> Cluster:
        """
        Sets this cluster layout's size to the specified two-element array of
        numbers :code:`[width, height]` and returns this cluster layout. A
        layout size of :code:`None` indicates that a node size will be used
        instead. The coordinates :code:`x` and :code:`y` represent an arbitrary
        coordinate system; for example, to produce a radial layout, a size of
        :code:`[360, radius]` corresponds to a breadth of 360° and a depth of
        radius.

        Parameters
        ----------
        size : tuple[float, float]
            Size values

        Returns
        -------
        Cluster
            Itself
        """
        self._node_size = False
        self._dx = size[0]
        self._dy = size[1]
        return self

    def set_node_size(self, size: tuple[float, float]) -> Cluster:
        """
        Sets this cluster layout's node size to the specified two-element array
        of numbers :code:`[width, height]` and returns this cluster layout. A
        node size of null indicates that a layout size will be used instead.
        When a :code:`size` is specified, the root node is always positioned at
        :math:`(0, 0)`.

        Parameters
        ----------
        size : tuple[float, float]
            Size values

        Returns
        -------
        Cluster
            Itself
        """
        self._node_size = True
        self._dx = size[0]
        self._dy = size[1]
        return self

    def get_separation(self) -> Callable[[Node, Node], int]:
        return self._separation

    def get_size(self) -> tuple[float, float] | None:
        if self._node_size:
            return None
        return [self._dx, self._dy]

    def get_node_size(self) -> tuple[float, float] | None:
        if self._node_size:
            return [self._dx, self._dy]
        return None


def cluster() -> Cluster:
    """
    Builds a new cluster layout with default settings.

    Returns
    -------
    Cluster
        Cluster object
    """
    return Cluster()



from math import sqrt

SQRT_3 = sqrt(3)


def symbol_asterisk(context, size):
    r = sqrt(size + min(size / 28, 0.75)) * 0.59436
    t = r / 2
    u = t * SQRT_3
    context.move_to(0, r)
    context.line_to(0, -r)
    context.move_to(-u, -t)
    context.line_to(u, t)
    context.move_to(-u, t)
    context.line_to(u, -t)



def tqmt(x: float) -> float:
    """
    Returns the two power minus ten times :code:`x` scaled to :math:`[0, 1]`.

    Parameters
    ----------
    x : float
        Input value

    Returns
    -------
    float
        Two power minus ten times the input value
    """
    return (pow(2, -10 * x) - 0.0009765625) * 1.0009775171065494



from math import sqrt


def ease_circle_in(t: float) -> float:
    """
    Circular easing.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_circle_in(0.2)
    0.020204102886728803
    """
    return 1 - sqrt(1 - t * t)


def ease_circle_out(t: float) -> float:
    """
    Reverse circular easing; equivalent to :code:`1 - d3.ease_circle_out(1 -
    t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_circle_out(0.2)
    0.5999999999999999
    """
    t -= 1
    return sqrt(1 - t * t)


def ease_circle_in_out(t: float) -> float:
    """
    Symmetric circle easing; scales :func:`d3.ease_circle_in
    <detroit.ease_circle_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_circle_out <detroit.ease_circle_out>` :math:`\\forall t \\in
    [0.5, 1]`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_circle_in_out(0.2)
    0.041742430504416006
    """
    t *= 2
    if t <= 1:
        return (1 - sqrt(1 - t * t)) * 0.5
    else:
        t -= 2
        return (sqrt(1 - t * t) + 1) * 0.5



from math import asin, cos, sin

from ...types import Point2D
from ..common import Projection
from .azimuthal import azimuthal_invert
from .projection import geo_projection

EPSILON = 1e-6


class OrthographicRaw:
    def __call__(self, x: float, y: float) -> Point2D:
        return [cos(y) * sin(x), sin(y)]

    def invert(self, x: float, y: float) -> Point2D:
        return azimuthal_invert(asin)(x, y)


def geo_orthographic() -> Projection:
    """
    The orthographic projection.

    Returns
    -------
    Projection
        Projection object
    """
    return geo_projection(OrthographicRaw()).scale(249.5).set_clip_angle(90 + EPSILON)



from math import asin, atan2, cos, degrees, pi, radians, sin

from ..types import Point2D
from .compose import Compose

TAU = 2 * pi


class RotationIdentity:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        if abs(lambda_) > pi:
            lambda_ -= round(lambda_ / TAU) * TAU
        return [lambda_, phi]

    def invert(self, lambda_: float, phi: float) -> Point2D:
        return self(lambda_, phi)


class RotateRadians:
    def __init__(self, delta_lambda: float, delta_phi: float, delta_gamma: float):
        delta_lambda %= TAU
        self._rotate = RotationIdentity()
        if delta_lambda:
            if delta_phi or delta_gamma:
                self._rotate = Compose(
                    RotationLambda(delta_lambda),
                    RotationPhiGamma(delta_phi, delta_gamma),
                )
            else:
                self._rotate = RotationLambda(delta_lambda)
        elif delta_phi or delta_gamma:
            self._rotate = RotationPhiGamma(delta_phi, delta_gamma)

    def __call__(self, lambda_: float, phi: float) -> Point2D:
        return self._rotate(lambda_, phi)

    def invert(self, lambda_: float, phi: float) -> Point2D:
        return self._rotate.invert(lambda_, phi)


class ForwardRotationLambda:
    def __init__(self, delta_lambda: float):
        self._delta_lambda = delta_lambda

    def __call__(self, lambda_: float, phi: float) -> Point2D:
        lambda_ += self._delta_lambda
        if abs(lambda_) > pi:
            lambda_ -= round(lambda_ / TAU) * TAU
        return [lambda_, phi]


class RotationLambda:
    def __init__(self, delta_lambda: float):
        self._call = ForwardRotationLambda(delta_lambda)
        self._invert = ForwardRotationLambda(-delta_lambda)

    def __call__(self, lambda_: float, phi: float) -> Point2D:
        return self._call(lambda_, phi)

    def invert(self, lambda_: float, phi: float) -> Point2D:
        return self._invert(lambda_, phi)


class RotationPhiGamma:
    def __init__(self, delta_phi: float, delta_gamma: float):
        self._cos_delta_phi = cos(delta_phi)
        self._sin_delta_phi = sin(delta_phi)
        self._cos_delta_gamma = cos(delta_gamma)
        self._sin_delta_gamma = sin(delta_gamma)

    def __call__(self, lambda_: float, phi: float) -> Point2D:
        cos_phi = cos(phi)
        x = cos(lambda_) * cos_phi
        y = sin(lambda_) * cos_phi
        z = sin(phi)
        k = z * self._cos_delta_phi + x * self._sin_delta_phi
        return [
            atan2(
                y * self._cos_delta_gamma - k * self._sin_delta_gamma,
                x * self._cos_delta_phi - z * self._sin_delta_phi,
            ),
            asin(k * self._cos_delta_gamma + y * self._sin_delta_gamma),
        ]

    def invert(self, lambda_: float, phi: float) -> Point2D:
        cos_phi = cos(phi)
        x = cos(lambda_) * cos_phi
        y = sin(lambda_) * cos_phi
        z = sin(phi)
        k = z * self._cos_delta_gamma - y * self._sin_delta_gamma
        return [
            atan2(
                y * self._cos_delta_gamma + z * self._sin_delta_gamma,
                x * self._cos_delta_phi + k * self._sin_delta_phi,
            ),
            asin(k * self._cos_delta_phi - x * self._sin_delta_phi),
        ]


class Rotation:
    """
    Rotation class
    """

    def __init__(self, rotate: tuple[float, float] | tuple[float, float, float]):
        self._rotate = RotateRadians(
            radians(rotate[0]),
            radians(rotate[1]),
            radians(rotate[2]) if len(rotate) > 2 else 0,
        )

    def __call__(self, coordinates: Point2D) -> Point2D:
        """
        Rotates a 2D point

        Parameters
        ----------
        coordinates : Point2D
            2D point

        Returns
        -------
        Point2D
            Rotated 2D point
        """
        coordinates = self._rotate(radians(coordinates[0]), radians(coordinates[1]))
        coordinates[0] = degrees(coordinates[0])
        coordinates[1] = degrees(coordinates[1])
        return coordinates

    def invert(self, coordinates: Point2D) -> Point2D:
        """
        Invert transformation.

        Parameters
        ----------
        coordinates : Point2D
            2D point

        Returns
        -------
        Point2D
            Inverted 2D point
        """
        coordinates = self._rotate.invert(
            radians(coordinates[0]), radians(coordinates[1])
        )
        coordinates[0] = degrees(coordinates[0])
        coordinates[1] = degrees(coordinates[1])
        return coordinates


def geo_rotation(rotate: tuple[float, float] | tuple[float, float, float]) -> Rotation:
    """
    Returns a rotation function for the given angles, which must be a two- or
    three-element array of numbers :math:`[\\lambda, \\phi, \\gamma]`
    specifying the rotation angles in degrees about each spherical axis. (These
    correspond to yaw, pitch and roll). If the rotation angle :math:`\\gamma`
    is omitted, it defaults to 0.


    Parameters
    ----------
    rotate : tuple[float, float] | tuple[float, float, float]
        Rotation parameters :math:`[\\lambda, \\phi, \\gamma]`

    Returns
    -------
    Rotation
        Rotation callable object
    """
    return Rotation(rotate)



from math import asin, atan2, cos, pi, sin, sqrt

from ...types import Point2D
from ..common import Projection, RawProjection
from .conic import conic_projection
from .cylindrical_equal_area import CylindricalEqualArea

EPSILON = 1e-6


def sign(x):
    return -1 if x < 0 else 1


class ConicEqualAreaRaw:
    def __init__(self, n: float, c: float):
        self._n = n
        self._c = c
        self._r0 = sqrt(c) / n

    def __call__(self, x: float, y: float) -> Point2D:
        r = sqrt(self._c - 2 * self._n * sin(y)) / self._n
        x *= self._n
        return [r * sin(x), self._r0 - r * cos(x)]

    def invert(self, x: float, y: float) -> Point2D:
        r0y = self._r0 - y
        l_ = atan2(x, abs(r0y)) * sign(r0y)
        if r0y * self._n < 0:
            l_ -= pi * sign(x) * sign(r0y)
        return [
            l_ / self._n,
            asin((self._c - (x * x + r0y * r0y) * self._n * self._n) / (2 * self._n)),
        ]


def conic_equal_area_raw(y0: float, y1: float) -> RawProjection:
    sy0 = sin(y0)
    n = (sy0 + sin(y1)) / 2
    if abs(n) < EPSILON:
        return CylindricalEqualArea(y0)
    c = 1 + sy0 * (2 * n - sy0)
    return ConicEqualAreaRaw(n, c)


def geo_conic_equal_area() -> Projection:
    """
    The Albers' equal-area conic projection.

    returns
    -------
    Projection
        Projection object
    """
    return (
        conic_projection(conic_equal_area_raw).scale(155.424).set_center([0, 33.6442])
    )



from math import atan, cos, fabs, nan, pi, sin

from ..common import LineStream
from .clip import clip

half_pi = pi * 0.5
EPSILON = 1e-6


class ClipAntimeridianLine(LineStream):
    def __init__(self, stream: LineStream):
        self._lambda0 = nan
        self._phi0 = nan
        self._sign0 = nan
        self._clean = None
        self._stream = stream

    def line_start(self):
        self._stream.line_start()
        self._clean = 1

    def line_end(self):
        self._stream.line_end()
        self._lambda0 = nan
        self._phi0 = nan

    def point(self, lambda1: float, phi1: float):
        sign1 = pi if lambda1 > 0 else -pi
        delta = fabs(lambda1 - self._lambda0)

        if abs(delta - pi) < EPSILON:
            self._phi0 = half_pi if (self._phi0 + phi1) * 0.5 > 0 else -half_pi
            self._stream.point(self._lambda0, self._phi0)
            self._stream.point(self._sign0, self._phi0)
            self._stream.line_end()
            self._stream.line_start()
            self._stream.point(sign1, self._phi0)
            self._stream.point(lambda1, self._phi0)
            self._clean = 0
        elif self._sign0 != sign1 and delta >= pi:
            if fabs(self._lambda0 - self._sign0) < EPSILON:
                self._lambda0 -= self._sign0 * EPSILON
            if fabs(lambda1 - sign1) < EPSILON:
                lambda1 -= sign1 * EPSILON
            self._phi0 = clip_antimeridian_intersect(
                self._lambda0, self._phi0, lambda1, phi1
            )
            self._stream.point(self._sign0, self._phi0)
            self._stream.line_end()
            self._stream.line_start()
            self._stream.point(sign1, self._phi0)
            self._clean = 0
        self._lambda0 = lambda1
        self._phi0 = phi1
        self._stream.point(self._lambda0, self._phi0)
        self._sign0 = sign1

    def clean(self) -> int:
        return 2 - self._clean

    def __str__(self):
        return f"ClipAntimeridianLine({self._stream})"


def clip_antimeridian_intersect(
    lambda0: float, phi0: float, lambda1: float, phi1: float
) -> float:
    sin_lambd0_lambda1 = sin(lambda0 - lambda1)
    if abs(sin_lambd0_lambda1) > EPSILON:
        cos_phi0 = cos(phi0)
        cos_phi1 = cos(phi1)
        return atan(
            (sin(phi0) * cos_phi1 * sin(lambda1) - sin(phi1) * cos_phi0 * sin(lambda0))
            / (cos_phi0 * cos_phi1 * sin_lambd0_lambda1)
        )
    else:
        return (phi0 + phi1) * 0.5


def clip_antimeridian_interpolate(
    vfrom: float | None,
    vto: float | None,
    direction: float,
    stream: LineStream,
):
    if vfrom is None:
        phi = direction * half_pi
        stream.point(-pi, phi)
        stream.point(0, phi)
        stream.point(pi, phi)
        stream.point(pi, 0)
        stream.point(pi, -phi)
        stream.point(0, -phi)
        stream.point(-pi, -phi)
        stream.point(-pi, 0)
        stream.point(-pi, phi)
    elif abs(vfrom[0] - vto[0]) > EPSILON:
        lambda_ = pi if vfrom[0] < vto[0] else -pi
        phi = direction * lambda_ * 0.5
        stream.point(-lambda_, phi)
        stream.point(0, phi)
        stream.point(lambda_, phi)
    else:
        stream.point(vto[0], vto[1])


def visible(*args) -> bool:
    return True


geo_clip_antimeridian = clip(
    visible, ClipAntimeridianLine, clip_antimeridian_interpolate, [-pi, -half_pi]
)
geo_clip_antimeridian.__doc__ = """
A clipping function which transforms a stream such that geometries (lines or
polygons) that cross the antimeridian line are cut in two, one on each side.
Typically used for pre-clipping.

Returns
-------
Callable[[PolygonStream], ClipRectangle]
    Clipping function
"""



from .binary import binary as treemap_binary
from .dice import dice as treemap_dice
from .resquarify import resquarify as treemap_resquarify
from .slice import slice as treemap_slice
from .slice_dice import slice_dice as treemap_slice_dice
from .squarify import squarify as treemap_squarify
from .treemap import treemap

__all__ = [
    "treemap",
    "treemap_binary",
    "treemap_dice",
    "treemap_resquarify",
    "treemap_slice",
    "treemap_slice_dice",
    "treemap_squarify",
]



class Polygon:
    def __init__(self):
        self._values = []

    def move_to(self, x: float, y: float):
        self._values.append([x, y])

    def close_path(self):
        self._values.append(self._values[0].copy())

    def line_to(self, x: float, y: float):
        self._values.append([x, y])

    def value(self) -> list[tuple[float, float]] | None:
        return self._values or None



from ..colors import colors
from ..ramp import ramp

scheme = [
    "ece2f0a6bddb1c9099",
    "f6eff7bdc9e167a9cf02818a",
    "f6eff7bdc9e167a9cf1c9099016c59",
    "f6eff7d0d1e6a6bddb67a9cf1c9099016c59",
    "f6eff7d0d1e6a6bddb67a9cf3690c002818a016450",
    "fff7fbece2f0d0d1e6a6bddb67a9cf3690c002818a016450",
    "fff7fbece2f0d0d1e6a6bddb67a9cf3690c002818a016c59014636",
]
SCHEME_PUBUGN = list(map(colors, scheme))
interpolate_pubugn = ramp(SCHEME_PUBUGN)



import math

degrees = 180 / math.pi

identity = {
    "translateX": 0,
    "translateY": 0,
    "rotate": 0,
    "skewX": 0,
    "scaleX": 1,
    "scaleY": 1,
}


def decompose(a, b, c, d, e, f):
    scaleX = math.sqrt(a * a + b * b)
    scaleY = math.sqrt(c * c + d * d)

    if scaleX:
        a /= scaleX
        b /= scaleX

    skewX = a * c + b * d
    if skewX:
        c -= a * skewX
        d -= b * skewX

    if scaleY:
        c /= scaleY
        d /= scaleY
        skewX /= scaleY

    if a * d < b * c:
        a, b, c, d = -a, -b, -c, -d
        scaleX, scaleY = -scaleX, -scaleY

    return {
        "translateX": e,
        "translateY": f,
        "rotate": math.atan2(b, a) * degrees,
        "skewX": math.atan(skewX) * degrees,
        "scaleX": scaleX,
        "scaleY": scaleY,
    }



from .categorical.accent import SCHEME_ACCENT
from .categorical.category10 import SCHEME_CATEGORY_10
from .categorical.dark2 import SCHEME_DARK_2
from .categorical.observable10 import SCHEME_OBSERVABLE_10
from .categorical.paired import SCHEME_PAIRED
from .categorical.pastel1 import SCHEME_PASTEL_1
from .categorical.pastel2 import SCHEME_PASTEL_2
from .categorical.set1 import SCHEME_SET_1
from .categorical.set2 import SCHEME_SET_2
from .categorical.set3 import SCHEME_SET_3
from .categorical.tableau10 import SCHEME_TABLEAU_10
from .diverging.br_bg import SCHEME_BRBG, interpolate_brbg
from .diverging.pi_yg import SCHEME_PIYG, interpolate_piyg
from .diverging.pr_gn import SCHEME_PRGN, interpolate_prgn
from .diverging.pu_or import SCHEME_PUOR, interpolate_puor
from .diverging.rd_bu import SCHEME_RDBU, interpolate_rdbu
from .diverging.rd_gy import SCHEME_RDGY, interpolate_rdgy
from .diverging.rd_yl_bu import SCHEME_RDYLBU, interpolate_rdylbu
from .diverging.rd_yl_gn import SCHEME_RDYLGN, interpolate_rdylgn
from .diverging.spectral import SCHEME_SPECTRAL, interpolate_spectral
from .sequential_multi.bu_gn import SCHEME_BUGN, interpolate_bugn
from .sequential_multi.bu_pu import SCHEME_BUPU, interpolate_bupu
from .sequential_multi.cividis import interpolate_cividis
from .sequential_multi.cubehelix import interpolate_cubehelix_default
from .sequential_multi.gn_bu import SCHEME_GNBU, interpolate_gnbu
from .sequential_multi.or_rd import SCHEME_ORRD, interpolate_orrd
from .sequential_multi.pu_bu import SCHEME_PUBU, interpolate_pubu
from .sequential_multi.pu_bu_gn import SCHEME_PUBUGN, interpolate_pubugn
from .sequential_multi.pu_rd import SCHEME_PURD, interpolate_purd
from .sequential_multi.rainbow import (
    interpolate_cool,
    interpolate_rainbow,
    interpolate_warm,
)
from .sequential_multi.rd_pu import SCHEME_RDPU, interpolate_rdpu
from .sequential_multi.sinebow import interpolate_sinebow
from .sequential_multi.turbo import interpolate_turbo
from .sequential_multi.viridis import (
    interpolate_inferno,
    interpolate_magma,
    interpolate_plasma,
    interpolate_viridis,
)
from .sequential_multi.yl_gn import SCHEME_YLGN, interpolate_ylgn
from .sequential_multi.yl_gn_bu import SCHEME_YLGNBU, interpolate_ylgnbu
from .sequential_multi.yl_or_br import SCHEME_YLORBR, interpolate_ylorbr
from .sequential_multi.yl_or_rd import SCHEME_YLORRD, interpolate_ylorrd
from .sequential_single.blues import SCHEME_BLUES, interpolate_blues
from .sequential_single.greens import SCHEME_GREENS, interpolate_greens
from .sequential_single.greys import SCHEME_GREYS, interpolate_greys
from .sequential_single.oranges import SCHEME_ORANGES, interpolate_oranges
from .sequential_single.purples import SCHEME_PURPLES, interpolate_purples
from .sequential_single.reds import SCHEME_REDS, interpolate_reds

__all__ = [
    "SCHEME_ACCENT",
    "SCHEME_BLUES",
    "SCHEME_BRBG",
    "SCHEME_BUGN",
    "SCHEME_BUPU",
    "SCHEME_CATEGORY_10",
    "SCHEME_DARK_2",
    "SCHEME_GNBU",
    "SCHEME_GREENS",
    "SCHEME_GREYS",
    "SCHEME_OBSERVABLE_10",
    "SCHEME_ORANGES",
    "SCHEME_ORRD",
    "SCHEME_PAIRED",
    "SCHEME_PASTEL_1",
    "SCHEME_PASTEL_2",
    "SCHEME_PIYG",
    "SCHEME_PRGN",
    "SCHEME_PUBU",
    "SCHEME_PUBUGN",
    "SCHEME_PUOR",
    "SCHEME_PURD",
    "SCHEME_PURPLES",
    "SCHEME_RDBU",
    "SCHEME_RDGY",
    "SCHEME_RDPU",
    "SCHEME_RDYLBU",
    "SCHEME_RDYLGN",
    "SCHEME_REDS",
    "SCHEME_SET_1",
    "SCHEME_SET_2",
    "SCHEME_SET_3",
    "SCHEME_SPECTRAL",
    "SCHEME_TABLEAU_10",
    "SCHEME_YLGN",
    "SCHEME_YLGNBU",
    "SCHEME_YLORBR",
    "SCHEME_YLORRD",
    "interpolate_blues",
    "interpolate_brbg",
    "interpolate_bugn",
    "interpolate_bupu",
    "interpolate_cividis",
    "interpolate_cool",
    "interpolate_cubehelix_default",
    "interpolate_gnbu",
    "interpolate_greens",
    "interpolate_greys",
    "interpolate_inferno",
    "interpolate_magma",
    "interpolate_oranges",
    "interpolate_orrd",
    "interpolate_piyg",
    "interpolate_plasma",
    "interpolate_prgn",
    "interpolate_pubu",
    "interpolate_pubugn",
    "interpolate_puor",
    "interpolate_purd",
    "interpolate_purples",
    "interpolate_rainbow",
    "interpolate_rdbu",
    "interpolate_rdgy",
    "interpolate_rdpu",
    "interpolate_rdylbu",
    "interpolate_rdylgn",
    "interpolate_reds",
    "interpolate_sinebow",
    "interpolate_spectral",
    "interpolate_turbo",
    "interpolate_viridis",
    "interpolate_warm",
    "interpolate_ylgn",
    "interpolate_ylgnbu",
    "interpolate_ylorbr",
    "interpolate_ylorrd",
]



from collections.abc import Callable
from math import asin, atan2, cos, isinf, sin, sqrt


def azimuthal_raw(
    scale: Callable[[float], float],
) -> Callable[[float, float], tuple[float, float]]:
    def call(x, y):
        cx = cos(x)
        cy = cos(y)
        k = scale(cx * cy)
        if isinf(k):
            return [2, 0]
        return [k * cy * sin(x), k * sin(y)]

    return call


def azimuthal_invert(
    angle: Callable[[float], float],
) -> Callable[[float, float], tuple[float, float]]:
    def invert(x, y):
        z = sqrt(x * x + y * y)
        c = angle(z)
        sc = sin(c)
        cc = cos(c)
        return [atan2(x * sc, z * cc), asin(y * sc / z if z != 0.0 else 0.0)]

    return invert



from ..types import Point2D


def cross(a: Point2D, b: Point2D, c: Point2D) -> float:
    """
    Returns the 2D cross product of AB and AC vectors, i.e., the z-component of
    the 3D cross product in a quadrant I Cartesian coordinate system (+x is
    right, +y is up). Returns a positive value if ABC is counter-clockwise,
    negative if clockwise, and zero if the points are collinear.

    Parameters
    ----------
    a : Point2D
        Point A
    b : Point2D
        Point B
    c : Point2D
        Point C

    Returns
    -------
    float
        2D cross product
    """
    return (b[0] - a[0]) * (c[1] - a[1]) - (b[1] - a[1]) * (c[0] - a[0])



from __future__ import annotations

from collections.abc import Callable
from math import sqrt

from ..accessors import optional
from ..constant import constant, constant_zero
from ..hierarchy import Node
from .siblings import pack_siblings_random


def default_radius(d: Node) -> float:
    return sqrt(d.value)


class Pack:
    """
    Pack layout
    """

    def __init__(self):
        self._radius = None
        self._dx = 1
        self._dy = 1
        self._padding = constant_zero

    def __call__(self, root: Node) -> Node:
        """
        Lays out the specified root hierarchy, assigning the following
        properties on root and its descendants:

        * :code:`node.x` - the x-coordinate of the circle’s center
        * :code:`node.y` - the y coordinate of the circle’s center
        * :code:`node.r` - the radius of the circle

        You must call root.sum before passing the hierarchy to the pack layout.
        You probably also want to call root.sort to order the hierarchy before
        computing the layout.


        Parameters
        ----------
        root : Node
            Root node

        Returns
        -------
        Node
            Node organized as pack
        """
        root.x = self._dx * 0.5
        root.y = self._dy * 0.5
        if self._radius:
            (
                root.each_before(radius_leaf(self._radius))
                .each_after(pack_children_random(self._padding, 0.5))
                .each_before(translate_child(1))
            )
        else:
            (
                root.each_before(radius_leaf(default_radius))
                .each_after(pack_children_random(constant_zero, 1))
                .each_after(
                    pack_children_random(
                        self._padding, root.r / min(self._dx, self._dy)
                    )
                )
                .each_before(translate_child(min(self._dx, self._dy) / (2 * root.r)))
            )
        return root

    def set_radius(self, radius: Callable[[Node], float] | None) -> Pack:
        """
        Sets the pack layout's radius accessor to the specified function and
        returns this pack layout. If :code:`radius` is not specified, returns
        the current radius accessor, which defaults to :code:`None`. If the
        radius accessor is :code:`None`, the radius of each leaf circle is
        derived from the leaf :code:`node.value` (computed by
        :code:`node.sum`); the radii are then scaled proportionally to fit the
        layout size. If the :code:`radius` accessor is not :code:`None`, the
        radius of each leaf circle is specified exactly by the function.

        Parameters
        ----------
        radius : Callable[[Node], float] | None
            Radius function or :code:`None` value

        Returns
        -------
        Pack
            Itself
        """
        self._radius = optional(radius)
        return self

    def set_size(self, size: tuple[float, float]) -> Pack:
        """
        Sets this pack layout's size to the specified two-element array of
        numbers :code:`[width, height]` and returns this pack layout.

        Parameters
        ----------
        size : tuple[float, float]
            Size values

        Returns
        -------
        Pack
            Itself
        """
        self._dx = size[0]
        self._dy = size[1]
        return self

    def set_padding(self, padding: Callable[[Node], float] | float) -> Pack:
        """
        If :code:`padding` is specified, sets this pack layout's padding
        accessor to the specified number or function and returns this pack
        layout. When siblings are packed, tangent siblings will be separated by
        approximately the specified padding; the enclosing parent circle will
        also be separated from its children by approximately the specified
        padding. If an explicit radius is not specified, the padding is
        approximate because a two-pass algorithm is needed to fit within the
        layout size: the circles are first packed without padding; a scaling
        factor is computed and applied to the specified padding; and lastly the
        circles are re-packed with padding.

        Parameters
        ----------
        padding : Callable[[Node], float] | float
            Padding function or constant padding value

        Returns
        -------
        Pack
            Itself
        """
        if callable(padding):
            self._padding = padding
        else:
            self._padding = constant(padding)
        return self

    def get_radius(self) -> Callable[[Node], None] | None:
        return self._radius

    def get_size(self) -> tuple[float, float]:
        return [self._dx, self._dy]

    def get_padding(self) -> Callable[[Node], None]:
        return self._padding


def radius_leaf(radius: Callable[[Node], float] | None) -> Callable[[Node], None]:
    def leaf(node: Node):
        if node.children is None:
            node.r = max(0, 0 if radius is None else radius(node))

    return leaf


def pack_children_random(
    padding: Callable[[Node], float], k: float
) -> Callable[[Node], None]:
    def pack_children(node: Node):
        if children := node.children:
            r = padding(node) * k
            if r:
                for child in children:
                    child.r += r
            e = pack_siblings_random(children)
            if r:
                for child in children:
                    child.r -= r
            node.r = e + r

    return pack_children


def translate_child(k: float) -> Callable[[Node], None]:
    def translate(node: Node):
        parent = node.parent
        node.r *= k
        if parent is not None:
            node.x = parent.x + k * node.x
            node.y = parent.y + k * node.y

    return translate


def pack() -> Pack:
    return Pack()



from .cluster import cluster
from .hierarchy import Node, hierarchy
from .pack import (
    pack,
    pack_enclose,
    pack_siblings,
)
from .partition import partition
from .stratify import stratify
from .tree import tree
from .treemap import (
    treemap,
    treemap_binary,
    treemap_dice,
    treemap_resquarify,
    treemap_slice,
    treemap_slice_dice,
    treemap_squarify,
)

__all__ = [
    "Node",
    "cluster",
    "hierarchy",
    "pack",
    "pack_enclose",
    "pack_siblings",
    "partition",
    "stratify",
    "tree",
    "treemap",
    "treemap_binary",
    "treemap_dice",
    "treemap_resquarify",
    "treemap_slice",
    "treemap_slice_dice",
    "treemap_squarify",
]



from math import inf

from ...types import T
from ..series import Series
from .none import order_none


def order_appearance(series: list[Series[T]]) -> list[int]:
    """
    Returns a series order such that the earliest series (according to the
    maximum value) is at the bottom.


    Parameters
    ----------
    series : list[Series[T]]
        List of series

    Returns
    -------
    list[int]
        List of ordered indices
    """
    peaks = list(map(peak, series))
    return sorted(order_none(series), key=lambda i: peaks[i])


def peak(series: Series) -> int:
    vj = -inf
    j = 0
    for i in range(len(series)):
        vi = float(series[i][1])
        if vi > vj:
            vj = vi
            j = i
    return j



B1 = 4 / 11
B2 = 6 / 11
B3 = 8 / 11
B4 = 3 / 4
B5 = 9 / 11
B6 = 10 / 11
B7 = 15 / 16
B8 = 21 / 22
B9 = 63 / 64
B0 = 1 / B1 / B1


def ease_bounce_in(t: float) -> float:
    """
    Bounce easing, like a rubber ball.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_bounce_in(0.2)
    0.06000000000000005
    """
    return 1 - ease_bounce_out(1 - t)


def ease_bounce_out(t: float) -> float:
    """
    Reverse bounce easing; equivalent to :code:`1 - ease_bounce_in(1 - t)`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_bounce_out(0.2)
    0.30250000000000005
    """
    if t < B1:
        return B0 * t * t
    elif t < B3:
        t -= B2
        return B0 * t * t + B4
    elif t < B6:
        t -= B5
        return B0 * t * t + B7
    else:
        t -= B8
        return B0 * t * t + B9


def ease_bounce_in_out(t: float) -> float:
    """
    Symmetric bounce easing; scales :func:`d3.ease_bounce_in
    <detroit.ease_bounce_in>` :math:`\\forall t \\in [0, 0.5]` and
    :func:`d3.ease_bounce_out <detroit.ease_bounce_out>` :math:`\\forall t \\in
    [0.5, 1]`.

    Parameters
    ----------
    t : float
        :math:`t \\in [0, 1]`

    Returns
    -------
    float
        Output value of the easing function

    Examples
    --------
    >>> d3.ease_bounce_in_out(0.2)
    0.11375000000000002
    """
    t *= 2
    if t <= 1:
        return (1 - ease_bounce_out(1 - t)) * 0.5
    else:
        return (ease_bounce_out(t - 1) + 1) * 0.5



# https://en.wikipedia.org/wiki/Linear_congruential_generator#Parameters_in_common_use
A = 1664525
C = 1013904223
M = 4294967296  # 2^32


class LCG:
    def __init__(self):
        self._s = 1

    def __call__(self):
        self._s = (A * self._s + C) % M
        return self._s / M


def lcg():
    return LCG()



from math import sqrt

C = -0.5
S = sqrt(3) / 2
K = 1 / sqrt(12)
A = (K / 2 + 1) * 3


def symbol_wye(context, size):
    r = sqrt(size / A)
    x0 = r / 2
    y0 = r * K
    x1 = x0
    y1 = r * K + r
    x2 = -x1
    y2 = y1
    context.move_to(x0, y0)
    context.line_to(x1, y1)
    context.line_to(x2, y2)
    context.line_to(C * x0 - S * y0, S * x0 + C * y0)
    context.line_to(C * x1 - S * y1, S * x1 + C * y1)
    context.line_to(C * x2 - S * y2, S * x2 + C * y2)
    context.line_to(C * x0 + S * y0, C * y0 - S * x0)
    context.line_to(C * x1 + S * y1, C * y1 - S * x1)
    context.line_to(C * x2 + S * y2, C * y2 - S * x2)
    context.close_path()



from __future__ import annotations

import math
from collections.abc import Callable
from typing import overload

from ..types import Number
from .continuous import Transformer, copy, identity
from .init import init_range
from .linear import LinearBase


def transform_pow(exponent: Number) -> Callable[[float], float]:
    return lambda x: (-math.pow(-x, exponent) if x < 0 else math.pow(x, exponent))


def transform_sqrt(x: float) -> float:
    return -math.sqrt(-x) if x < 0 else math.sqrt(x)


def transform_square(x: float) -> float:
    return -x * x if x < 0 else x * x


class ScalePow(Transformer[float], LinearBase):
    """
    Power ("pow") scales are similar to linear scales, except an exponential
    transform is applied to the input domain value before the output range
    value is computed.
    Each range value y can be expressed as a function of the domain value x:
    :math:`y = m \\cdot x^k + b`, where :math:`k` is the exponent value.
    Power scales also support negative domain values, in which case the
    input value and the resulting output value are multiplied by -1.

    Parameters
    ----------
    t : Callable[[float], float]
        Transform function
    u : Callable[[float], float]
        Untransform function
    exponent : float | int
        Exponent
    """

    def __init__(
        self,
        t: Callable[[float], float] = identity,
        u: Callable[[float], float] = identity,
        exponent: float | int = 1,
    ):
        super().__init__(t, u)
        self._exponent = exponent

    def _pow_rescale(self):
        if self._exponent == 1:
            self._transform = identity
            self._untransform = identity
            self._rescale()
            return self
        elif self._exponent == 0.5:
            self._transform = transform_sqrt
            self._untransform = transform_square
            self._rescale()
            return self
        else:
            self._transform = transform_pow(self._exponent)
            self._untransform = transform_pow(1 / self._exponent)
            self._rescale()
            return self

    def set_exponent(self, exponent: Number) -> ScalePow:
        """
        Sets the current exponent to the given numeric value.

        Parameters
        ----------
        exponent : Number
            exponent

        Returns
        -------
        ScalePow
            Itself
        """
        self._exponent = float(exponent)
        return self._pow_rescale()

    def get_exponent(self) -> Number:
        return self._exponent

    def copy(self):
        return copy(self, ScalePow()).set_exponent(self.get_exponent())


@overload
def scale_pow() -> ScalePow: ...


@overload
def scale_pow(range_vals: list[float]) -> ScalePow: ...


@overload
def scale_pow(domain: list[Number], range_vals: list[float]) -> ScalePow: ...


def scale_pow(*args) -> ScalePow:
    """
    Constructs a new pow scale with the specified domain and
    range, the exponent 1, the default interpolator and
    clamping disabled.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[float]
        Array of values

    Returns
    -------
    ScalePow
        Scale object

    Examples
    --------
    >>> scale = d3.scale_pow([0, 10], [0, 960])
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     print(x, scale(x))
    ...
    ...
    0 0.0
    1 96.0
    2 192.0
    3 288.0
    4 384.0
    5 480.0
    6 576.0
    7 672.0
    8 768.0
    9 864.0
    10 960.0

    With :code:`exponent = 2`:

    >>> scale = scale.set_exponent(2)
    >>> for x in range(steps + 1):
    ...     print(x, scale(x))
    ...
    ...
    0 0.0
    1 9.6
    2 38.4
    3 86.39999999999999
    4 153.6
    5 240.0
    6 345.59999999999997
    7 470.4
    8 614.4
    9 777.6
    10 960.0
    """
    scale = ScalePow()
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)


@overload
def scale_sqrt() -> ScalePow: ...


@overload
def scale_sqrt(range_vals: list[float]) -> ScalePow: ...


@overload
def scale_sqrt(domain: list[Number], range_vals: list[float]) -> ScalePow: ...


def scale_sqrt(*args) -> ScalePow:
    """
    Constructs a new pow scale with the specified domain and range,
    the exponent 0.5, the default interpolator and clamping disabled.

    Parameters
    ----------
    domain : list[Number]
        Array of numbers
    range_vals : list[float]
        Array of values

    Returns
    -------
    ScalePow
        Scale object

    Examples
    --------
    >>> scale = d3.scale_sqrt([0, 10], [0, 960])
    >>> steps = 10
    >>> for x in range(steps + 1):
    ...     print(x, scale(x))
    ...
    ...
    0 0.0
    1 303.5786553761644
    2 429.3250516799596
    3 525.8136552049594
    4 607.1573107523288
    5 678.8225099390855
    6 743.612802471824
    7 803.1936254727125
    8 858.6501033599192
    9 910.7359661284933
    10 960.0
    """
    scale = ScalePow().set_exponent(0.5)
    if len(args) == 1:
        return init_range(scale, range_vals=args[0])
    elif len(args) == 2:
        domain, range_vals = args
        return init_range(scale, domain=domain, range_vals=range_vals)
    return init_range(scale)



from __future__ import annotations

from collections.abc import Iterator

from ..hierarchy import Node
from .dice import dice
from .slice import slice
from .squarify import PHI, squarify_ratio


class SquarifyList:
    def __init__(self, values: list[Node]):
        self._values = values
        self.ratio = None

    def __len__(self) -> int:
        return len(self._values)

    def __iter__(self) -> Iterator[Node]:
        return iter(self._values)

    def __setitem__(self, index: int, value: Node):
        self._values[index] = value

    def __getitem__(self, index: int) -> Node:
        return self._values[index]


class Resquarify:
    """
    Resquarify object

    Parameters
    ----------
    ratio : float
        Ratio value
    """

    def __init__(self, ratio: float):
        self._ratio = ratio

    def __call__(self, parent: Node, x0: float, y0: float, x1: float, y1: float):
        """
        Like :func:`d3.treemap_resquarify <detroit.resquarify>`, except
        preserves the topology (node adjacencies) of the previous layout
        computed by this function, if there is one and it used the same target
        aspect ratio. This tiling method is good for animating changes to
        treemaps because it only changes node sizes and not their relative
        positions, thus avoiding distracting shuffling and occlusion. The
        downside of a stable update, however, is a suboptimal layout for
        subsequent updates: only the first layout uses the Bruls et al.
        squarified algorithm.

        Parameters
        ----------
        parent : Node
            Parent node
        x0 : float
            X-coordinate rectangular edge
        y0 : float
            Y-coordinate rectangular edge
        x1 : float
            X-coordinate rectangular edge
        y1 : float
            Y-coordinate rectangular edge
        """
        rows = parent._squarify if hasattr(parent, "_squarify") else None
        if rows and rows.ratio == self._ratio:
            j = 0
            m = len(rows)
            value = parent.value

            while j < m:
                row = rows[j]
                nodes = row.children
                row.value = sum((node.value for node in nodes))
                if row.dice:
                    if value:
                        y = y0 + (y1 - y0) * row.value / value
                        dice(row, x0, y0, x1, y)
                        y0 = y
                    else:
                        dice(row, x0, y0, x1, y1)
                else:
                    if value:
                        x = x0 + (x1 - x0) * row.value / value
                        slice(row, x0, y0, x, y1)
                        x0 = x
                    else:
                        slice(row, x0, y0, x1, y1)
                value -= row.value
                j += 1
        else:
            parent._squarify = rows = SquarifyList(
                squarify_ratio(self._ratio, parent, x0, y0, x1, y1)
            )
            rows.ratio = self._ratio

    def set_ratio(self, ratio: float) -> Resquarify:
        """
        Specifies the desired aspect ratio of the generated rectangles. The
        ratio must be specified as a number greater than or equal to one. Note
        that the orientation of the generated rectangles (tall or wide) is not
        implied by the ratio; for example, a ratio of two will attempt to
        produce a mixture of rectangles whose width:height ratio is either
        :code:`2:1` or :code:`1:2`. (However, you can approximately achieve
        this result by generating a square treemap at different dimensions, and
        then stretching the treemap to the desired aspect ratio.) Furthermore,
        the specified ratio is merely a hint to the tiling algorithm; the
        rectangles are not guaranteed to have the specified aspect ratio. If
        not specified, the aspect ratio defaults to the golden ratio,
        :math:`\\phi = (1 + \\sqrt{5}) / 2`, per `Kong et
        al<http://vis.stanford.edu/papers/perception-treemaps>_`.

        Parameters
        ----------
        ratio : float
            Ratio value

        Returns
        -------
        Resquarify
            New Resquarify object
        """
        return Resquarify(ratio if ratio > 1 else 1)


resquarify = Resquarify(PHI)



from ..colors import colors
from ..ramp import ramp

scheme = [
    "fee0d2fc9272de2d26",
    "fee5d9fcae91fb6a4acb181d",
    "fee5d9fcae91fb6a4ade2d26a50f15",
    "fee5d9fcbba1fc9272fb6a4ade2d26a50f15",
    "fee5d9fcbba1fc9272fb6a4aef3b2ccb181d99000d",
    "fff5f0fee0d2fcbba1fc9272fb6a4aef3b2ccb181d99000d",
    "fff5f0fee0d2fcbba1fc9272fb6a4aef3b2ccb181da50f1567000d",
]
SCHEME_REDS = list(map(colors, scheme))
interpolate_reds = ramp(SCHEME_REDS)



from .format_decimal import format_decimal
from .format_prefix_auto import format_prefix_auto
from .format_rounded import format_rounded

format_types = {
    "%": lambda x, p: f"{x * 100:.{p}f}",
    "b": lambda x: bin(round(x))[2:],
    "c": lambda x: str(x),
    "d": format_decimal,
    "e": lambda x, p: f"{x:.{p}e}",
    "f": lambda x, p: f"{x:.{p}f}",
    "g": lambda x, p: f"{x:.{p}g}",
    "o": lambda x: oct(round(x))[2:],
    "p": lambda x, p: format_rounded(x * 100, p),
    "r": format_rounded,
    "s": format_prefix_auto,
    "X": lambda x: hex(round(x))[2:].upper(),
    "x": lambda x: hex(round(x))[2:],
}



import math
from typing import TypeVar, overload

from .color import BRIGHTER, DARKER, RGB, Color, rgb_convert
from .lab import HCL, hcl_convert

TCubehelix = TypeVar("Cubehelix", bound="Cubehelix")

A = -0.14861
B = +1.78277
C = -0.29227
D = -0.90649
E = +1.97294
ED = E * D
EB = E * B
BC_DA = B * C - D * A


def cubehelix_convert(obj):
    if isinstance(obj, Cubehelix):
        return Cubehelix(obj.h, obj.s, obj.l, obj.opacity)
    if not isinstance(obj, RGB):
        obj = rgb_convert(obj)
    r = obj.r / 255
    g = obj.g / 255
    b = obj.b / 255
    l = (BC_DA * b + ED * r - EB * g) / (BC_DA + ED - EB)
    bl = b - l
    k = (E * (g - l) - C * bl) / D
    if l == 0 or l == 1:
        s = math.nan
    else:
        s = (k * k + bl * bl) ** 0.5 / (E * l * (1 - l))  # NaN if l=0 or l=1
    h = (math.degrees(math.atan2(k, bl)) - 120) if not math.isnan(s) and s else math.nan
    return Cubehelix(h + 360 if h < 0 else h, s, l, obj.opacity)


def lch(*args):
    if len(args) == 1:
        return hcl_convert(args[0])
    elif len(args) == 3:
        l, c, h = args
        opacity = 1
        return HCL(h, c, l, opacity)
    elif len(args) == 4:
        l, c, h, opacity = args
        return HCL(h, c, l, opacity)


class Cubehelix(Color):
    """
    Cubehelix color format

    Parameters
    ----------
    h : float
        Hue channel value
    s : float
        Saturation channel value
    l : float
        Lightness channel value
    opacity : float
        Opacity value
    """

    def __init__(self, h: float, s: float, l: float, opacity: float = 1.0):
        self.h = float(h)
        self.s = float(s)
        self.l = float(l)
        self.opacity = float(opacity)

    def brighter(self, k: float | None = None) -> TCubehelix:
        """
        Returns a brighter copy of this color.

        Parameters
        ----------
        k : float | None
            Brightness coefficient

        Returns
        -------
        Cubehelix
            Brighter Cubehelix
        """
        k = BRIGHTER if k is None else BRIGHTER**k
        return Cubehelix(self.h, self.s, self.l * k, self.opacity)

    def darker(self, k: float | None = None) -> TCubehelix:
        """
        Returns a darker copy of this color.

        Parameters
        ----------
        k : float | None
            Darkness coefficient

        Returns
        -------
        Cubehelix
            Darker Cubehelix
        """
        k = DARKER if k is None else DARKER**k
        return Cubehelix(self.h, self.s, self.l * k, self.opacity)

    def rgb(self) -> RGB:
        """
        Returns the RGB equivalent of this color

        Returns
        -------
        RGB
            RGB color format
        """
        h = 0 if math.isnan(self.h) else math.radians(self.h + 120)
        l = self.l
        a = 0 if math.isnan(self.s) else self.s * l * (1 - l)
        cosh = math.cos(h)
        sinh = math.sin(h)
        return RGB(
            255 * (l + a * (A * cosh + B * sinh)),
            255 * (l + a * (C * cosh + D * sinh)),
            255 * (l + a * (E * cosh)),
            self.opacity,
        )

    def __repr__(self) -> str:
        return f"Cubehelix(h={self.h}, s={self.s}, l={self.l}, opacity={self.opacity})"


@overload
def cubehelix(specifier: str) -> Cubehelix: ...


@overload
def cubehelix(h: float, s: float, l: float) -> Cubehelix: ...


@overload
def cubehelix(h: float, s: float, l: float, opacity: float) -> Cubehelix: ...


def cubehelix(*args):
    """
    Build a new Cubehelix color

    Parameters
    ----------
    specifier : str
        String which represents a color
    h : float
        Hue channel value between 0 and 360
    s : float
        Saturation channel value between 0 and 1
    l : float
        Lightness channel value between 0 and 1
    opacity : float
        Opacity value between 0 and 1

    Returns
    -------
    Cubehelix
        Cubehelix object

    Examples
    --------

    >>> d3.cubehelix("#2e65ffff")
    Cubehelix(h=222.45389380826435, s=1.3363126255145055, l=0.3978037691833379, opacity=1.0)
    >>> d3.cubehelix(210.4, 0.9, 0.7, 0.8)
    Cubehelix(h=210.4, s=0.9, l=0.7, opacity=0.8)
    """
    if len(args) == 1:
        return cubehelix_convert(args[0])
    elif len(args) == 3:
        h, s, l = args
        opacity = 1
        return Cubehelix(h, s, l, opacity)
    elif len(args) == 4:
        h, s, l, opacity = args
        return Cubehelix(h, s, l, opacity)



from __future__ import annotations

from collections.abc import Callable
from math import floor, isnan
from typing import Any

from ...types import GeoJSON, Point2D
from ..common import Context, Projection
from ..stream import geo_stream
from .area import AreaStream
from .bounds import BoundsStream
from .centroid import CentroidStream
from .context import PathContext
from .measure import LengthStream
from .string import PathString


def identity(x):
    return x


class GeoPath:
    """
    Creates a new geographic path generator with the default settings. If
    projection is specified, sets the current projection. If context is
    specified, sets the current context.

    Parameters
    ----------
    projection : Projection | None
        Projection
    context : Context | None
        Context
    """

    def __init__(
        self, projection: Projection | None = None, context: Context | None = None
    ):
        self._digits = 3
        self._point_radius = 4.5
        self._projection = projection
        self._projection_stream = identity if projection is None else projection.stream
        self._context = context
        self._context_stream = (
            PathString(self._digits) if context is None else PathContext(context)
        )

    def __call__(self, obj: GeoJSON, *args: Any) -> str | None:
        """
        Renders the given object, which may be any GeoJSON feature or geometry
        object:

        * :code:`Point` - a single position
        * :code:`MultiPoint` - an array of positions
        * :code:`LineString` - an array of positions forming a continuous line
        * :code:`MultiLineString` - an array of arrays of positions forming \
        several lines
        * :code:`Polygon` - an array of arrays of positions forming a polygon \
        (possibly with holes)
        * :code:`MultiPolygon` - a multidimensional array of positions forming \
        multiple polygons
        * :code:`GeometryCollection` - an array of geometry objects
        * :code:`Feature` - a feature containing one of the above geometry \
        objects
        * :code:`FeatureCollection` - an array of feature objects

        The type Sphere is also supported, which is useful for rendering the
        outline of the globe; a sphere has no coordinates. Any additional
        arguments are passed along to the pointRadius accessor.

        Parameters
        ----------
        obj : GeoJSON
            GeoJSON object
        *args : Any
            Additional arguments passed to :code:`point_radius` function

        Returns
        -------
        str | None
            Path
        """
        if obj:
            if callable(self._point_radius):
                self._context_stream.point_radius(self._point_radius(*args))
            geo_stream(obj, self._projection_stream(self._context_stream))

        return self._context_stream.result()

    def area(self, obj: GeoJSON) -> float:
        """
        Returns the projected planar area (typically in square pixels) for the
        specified GeoJSON object.

        Parameters
        ----------
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        float
            Area in px²
        """
        path_area = AreaStream()
        geo_stream(obj, self._projection_stream(path_area))
        return path_area.result()

    def measure(self, obj: GeoJSON) -> float:
        """
        Returns the projected planar length (typically in pixels) for the
        specified GeoJSON object.

        Parameters
        ----------
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        float
            Projected planar length in px
        """
        path_measure = LengthStream()
        geo_stream(obj, self._projection_stream(path_measure))
        return path_measure.result()

    def bounds(self, obj: GeoJSON) -> tuple[Point2D, Point2D]:
        """
        Returns the projected planar bounding box (typically in pixels) for the
        specified GeoJSON object.

        Parameters
        ----------
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        tuple[Point2D, Point2D]
            Projected planar bounding box
        """
        path_bounds = BoundsStream()
        geo_stream(obj, self._projection_stream(path_bounds))
        return path_bounds.result()

    def centroid(self, obj: GeoJSON) -> Point2D:
        """
        Returns the projected planar centroid (typically in pixels) for the
        specified GeoJSON object.

        Parameters
        ----------
        obj : GeoJSON
            GeoJSON object

        Returns
        -------
        Point2D
            Projected planar centroid
        """
        path_centroid = CentroidStream()
        geo_stream(obj, self._projection_stream(path_centroid))
        return path_centroid.result()

    def set_projection(self, projection: Projection | None = None) -> GeoPath:
        """
        If a projection is specified, sets the current projection to the
        specified projection.
        The projection defaults to :code:`None`, which represents the identity
        transformation: the input geometry is not projected and is instead
        rendered directly in raw coordinates. This can be useful for fast
        rendering of pre-projected geometry, or for fast rendering of the
        equirectangular projection.

        The given projection is typically one of built-in geographic
        projections; however, any object that exposes a projection.stream
        function can be used, enabling the use of custom projections.

        Parameters
        ----------
        projection : Projection | None
            Projection

        Returns
        -------
        GeoPath
            Itself
        """
        if projection is None:
            self._projection = None
            self._projection_stream = identity
        else:
            self._projection = projection
            self._projection_stream = projection.stream
        return self

    def set_context(self, context: Context | None = None) -> GeoPath:
        """
        If context is specified, sets the current render context and returns
        the path generator. If the context is :code:`None`, then the path
        generator will return an SVG path string; if the context is not
        :code:`None`, the path generator will instead call methods on the
        specified context to render geometry.

        Parameters
        ----------
        context : Context | None
            Context

        Returns
        -------
        GeoPath
            Itself
        """
        if context is None:
            self._context = None
            self._context_stream = PathString(self._digits)
        else:
            self._context = context
            self._context_stream = PathContext(context)
        return self

    def set_point_radius(self, radius: Callable[..., float] | float) -> GeoPath:
        """
        If radius is specified, sets the radius used to display Point and
        MultiPoint geometries to the specified number.

        Parameters
        ----------
        radius : Callable[..., float] | float
            Radius function or constant value

        Returns
        -------
        GeoPath
            Itself
        """
        if callable(radius):
            self._point_radius = radius
        else:
            self._context_stream.point_radius(radius)
            self._point_radius = radius
        return self

    def set_digits(self, digits: str | float | None = None) -> GeoPath:
        """
        If digits is specified (as a non-negative number), sets the number of
        fractional digits for coordinates generated in SVG path strings.

        Parameters
        ----------
        digits : str | float | None
            Digits value

        Returns
        -------
        GeoPath
            Itself
        """
        if digits is None:
            self._digits = None
        else:
            d = floor(float(digits)) if digits else 0
            if isnan(d) or d < 0:
                raise ValueError(f"Invalid digits: {digits}")
            self._digits = d
        if self._context is None:
            self._context_stream = PathString(self._digits)
        return self

    def get_projection(self):
        return self._projection

    def get_context(self):
        return self._context

    def get_point_radius(self):
        return self._point_radius

    def get_digits(self):
        return self._digits



from datetime import datetime, timedelta

from .interval import TimeInterval


class TimeMonth(TimeInterval):
    """
    Months in local time; ranges from 28 days to 31 days
    """

    def floor(self, date: datetime) -> datetime:
        """
        Returns a new date representing the latest interval
        boundary date before or equal to date.

        Parameters
        ----------
        date : datetime
            Date

        Returns
        -------
        datetime
            Floored date
        """
        return date.replace(day=1, hour=0, minute=0, second=0, microsecond=0)

    def offset(self, date: datetime, step: int) -> datetime:
        """
        Returns a new date equal to date plus step intervals.

        Parameters
        ----------
        date : datetime
            Date
        step : int
            Step to add or substitute

        Returns
        -------
        datetime
            Date
        """
        return date + timedelta(days=step * 31)

    def count(self, start: datetime, end: datetime) -> int:
        """
        Returns the number of interval boundaries after start
        (exclusive) and before or equal to end (inclusive).

        Parameters
        ----------
        start : datetime
            Start date
        end : datetime
            End date

        Returns
        -------
        int
            Count
        """
        days = (end - start).days
        return days // 30 + bool(0 < days % 30 <= 15)

    def field(self, date: datetime) -> int:
        """
        Returns the field value of the specified date, corresponding to the
        number of boundaries between this date (exclusive) and the latest
        previous parent boundary.

        Parameters
        ----------
        date : datetime
            Floored date

        Returns
        -------
        int
            Field value
        """
        return date.month - 1


time_month = TimeMonth()
time_months = time_month.range



from math import atan2, cos, pi, sin, sqrt

from ...types import Point2D
from ..common import Projection, RawProjection
from .conic import conic_projection
from .equirectangular import EquirectangularRaw

EPSILON = 1e-6


def sign(x):
    return -1 if x < 0 else 1


class ConicEquidistantRaw:
    def __init__(self, n: float, g: float):
        self._n = n
        self._g = g

    def __call__(self, x: float, y: float) -> Point2D:
        gy = self._g - y
        nx = self._n * x
        return [gy * sin(nx), self._g - gy * cos(nx)]

    def invert(self, x: float, y: float) -> Point2D:
        gy = self._g - y
        l_ = atan2(x, abs(gy)) * sign(gy)
        if gy * self._n < 0:
            l_ -= pi * sign(x) * sign(gy)
        return [l_ / self._n, self._g - sign(self._n) * sqrt(x * x + gy * gy)]


def conic_equidistant_raw(y0: float, y1: float) -> RawProjection:
    cy0 = cos(y0)
    n = sin(y0) if y0 == y1 else (cy0 - cos(y1)) / (y1 - y0)
    g = cy0 / n + y0
    return EquirectangularRaw() if abs(n) < EPSILON else ConicEquidistantRaw(n, g)


def geo_conic_equidistant() -> Projection:
    """
    The Albers' equidistant conic projection.

    returns
    -------
    Projection
        Projection object
    """
    return (
        conic_projection(conic_equidistant_raw).scale(131.154).set_center([0, 13.9389])
    )



import math

from .constant import constant


def linear(a, d):
    def f(t):
        return a + t * d

    return f


def exponential(a, b, y):
    a = a**y
    b = b**y - a
    y = 1 / y

    def f(t):
        return (a + t * b) ** y

    return f


def hue(a, b):
    d = b - a
    if not math.isnan(d) and d:
        if d < -180 or 180 < d:
            return linear(a, d - 360 * round(d / 360))
        else:
            return linear(a, d)
    else:
        return constant(b if math.isnan(a) else a)


def gamma(y):
    if round(y) == 1.0:
        return color

    def f(a, b):
        d = b - a
        return (
            exponential(a, b, y)
            if not math.isnan(d) and d
            else constant(math.isnan(a) and b or a)
        )

    return f


def color(a, b):
    d = b - a
    return (
        linear(a, d) if not math.isnan(d) and d else constant(b if math.isnan(a) else a)
    )



from lxml import etree

from .namespace import namespace
from .selection import Selection


def create(name: str) -> Selection:
    """
    Given the specified element name, returns a single-element selection
    containing a detached element of the given name in the current document.
    This method assumes the HTML namespace, so you must specify a namespace
    explicitly when creating SVG or other non-HTML elements.

    Parameters
    ----------
    name : str
        Tag name

    Returns
    -------
    Selection
        XML tree
    """
    fullname = namespace(name)
    document = (
        etree.Element(fullname["local"], attrib={"xmlns": fullname["space"]})
        if isinstance(fullname, dict)
        else etree.Element(fullname)
    )
    return Selection([[document]], [document])



from math import sqrt


def symbol_plus(context, size):
    r = sqrt(size - min(size / 7, 2)) * 0.87559
    context.move_to(-r, 0)
    context.line_to(r, 0)
    context.move_to(0, r)
    context.line_to(0, -r)



from collections.abc import Callable

from ..array import argpass
from ..types import Accessor, T


def matcher(match: Accessor[T, bool] | int | float | str) -> Callable[[T], bool]:
    """
    Returns a function whichs checks if the data input matches a arbitrary condition.

    Parameters
    ----------
    match : Accessor[T, bool] | int | float | str
        Match function or constant that data input should match.

    Returns
    -------
    Callable[[T], bool]
        Matching function
    """
    if callable(match):
        return argpass(match)

    def match_func(d: T) -> bool:
        return d == match

    return argpass(match_func)



from __future__ import annotations

from collections.abc import Callable
from math import cosh, dist, exp, log, sinh, sqrt, tanh

EPSILON = 1e-12
SQRT2 = sqrt(2)


class Base:
    def __init__(self, rho, p0, p1):
        ux0, uy0, w0 = p0
        ux1, uy1, w1 = p1
        dx = ux1 - ux0
        dy = uy1 - uy0
        self.rho = rho
        self.ux0 = ux0
        self.uy0 = uy0
        self.w0 = w0
        self.w1 = w1
        self.dx = dx
        self.dy = dy
        self.d2 = dx * dx + dy * dy


class SingularZoomRhoInterpolator(Base):
    def __init__(self, rho, rho2, rho4, p0, p1):
        super().__init__(rho, p0, p1)
        w0 = self.w0
        w1 = self.w1
        self.s = log(w1 / w0) / rho
        self.duration = self.s * 1000 * rho / SQRT2

    def __call__(self, t):
        return [
            self.ux0 + t * self.dx,
            self.uy0 + t * self.dy,
            self.w0 * exp(self.rho * t * self.s),
        ]


class GeneralZoomRhoInterpolator(Base):
    def __init__(self, rho, rho2, rho4, p0, p1):
        super().__init__(rho, p0, p1)
        d2 = self.d2
        w0 = self.w0
        w1 = self.w1

        self.rho2 = rho2
        self.d1 = d1 = sqrt(d2)

        b0 = (w1 * w1 - w0 * w0 + rho4 * d2) / (2 * w0 * rho2 * d1)
        b1 = (w1 * w1 - w0 * w0 - rho4 * d2) / (2 * w1 * rho2 * d1)
        self.r0 = r0 = log(sqrt(b0 * b0 + 1) - b0)
        self.r1 = r1 = log(sqrt(b1 * b1 + 1) - b1)

        self.s = (r1 - r0) / rho
        self.duration = self.s * 1000 * self.rho / SQRT2

    def __call__(self, t):
        c = self.w0 / (self.rho2 * self.d1)
        u = c * (cosh(self.r0) * tanh(self.rho * t * self.s + self.r0) - sinh(self.r0))
        return [
            self.ux0 + u * self.dx,
            self.uy0 + u * self.dy,
            self.w0 * cosh(self.r0) / cosh(self.rho * t * self.s + self.r0),
        ]


class ZoomRho:
    """
    An interpolator for zooming smoothly between two views of a two-dimensional
    plane based on "Smooth and efficient zooming and panning" by Jarke J. van
    Wijk and Wim A.A. Nuij.

    Each view is defined as an array of three numbers: cx, cy and width. The
    first two coordinates cx, cy represent the center of the viewport; the last
    coordinate width represents the size of the viewport.


    Parameters
    ----------
    a : list[float]
        View a
    b : list[float]
        View b

    Returns
    -------
    Callable[[float], float]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_zoom([0, 0, 1], [10, 10, 5])
    >>> interpolator(0)
    [0.0, 0.0, 1.0]
    >>> interpolator(1)
    [10.00000000000023, 10.00000000000023, 5.0000000000001155]
    >>> interpolator(0.5)
    [1.666666666666673, 1.666666666666673, 10.775486583496573]
    """

    def __init__(self, rho, rho2, rho4):
        self.rho = rho
        self.rho2 = rho2
        self.rho4 = rho4

    def __call__(self, a: list[float], b: list[float]) -> Callable[[float], float]:
        """
        Returns an interpolator between the two views a and b. Each view is
        defined as an array of three numbers: cx, cy and width. The first two
        coordinates cx, cy represent the center of the viewport; the last
        coordinate width represents the size of the viewport.

        Parameters
        ----------
        a : list[float]
            View a
        b : list[float]
            View b

        Returns
        -------
        Callable[[float], float]
            Interpolator function
        """
        d = dist(a[:-1], b[:-1]) ** 2
        interpolator = (
            SingularZoomRhoInterpolator if d < EPSILON else GeneralZoomRhoInterpolator
        )
        return interpolator(self.rho, self.rho2, self.rho4, a, b)

    def set_rho(self, rho: float) -> ZoomRho:
        """
        Sets rho value and returns itself. When :code:`rho` is closed to 0, the
        interpolator is almost linear.

        Parameters
        ----------
        rho : float
            Rho value

        Returns
        -------
        ZoomRho
            Itself
        """
        new_rho = max(1e-3, float(rho))
        return ZoomRho(new_rho, new_rho * new_rho, new_rho**4)


interpolate_zoom = ZoomRho(SQRT2, 2, 4)



from __future__ import annotations

from collections.abc import Callable

from ..selection import Selection
from ..types import Accessor, Number, T
from .area import Area
from .curves import Curve, curve_radial, curve_radial_linear
from .line_radial import LineRadial


class AreaRadial(Area[T]):
    def __init__(self):
        Area.__init__(self)
        super().set_curve(curve_radial_linear)

    def set_curve(self, curve: Callable[[Selection], Curve]) -> AreaRadial:
        """
        Sets curve.

        Parameters
        ----------
        curve : Callable[[Selection], Curve]
            Curve factory function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().set_curve(curve_radial(curve))

    def angle(self, angle: Accessor[T, float] | Number) -> AreaRadial:
        """
        Sets angle accessor function

        Parameters
        ----------
        angle : Accessor[T, float] | Number
            angle accessor function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().x(angle)

    def start_angle(self, start_angle: Accessor[T, float] | Number) -> AreaRadial:
        """
        Sets start angle accessor function

        Parameters
        ----------
        start_angle : Accessor[T, float] | Number
            angle accessor function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().x0(start_angle)

    def end_angle(self, end_angle: Accessor[T, float] | Number) -> AreaRadial:
        """
        Sets start angle accessor function

        Parameters
        ----------
        end_angle : Accessor[T, float] | Number
            angle accessor function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().x1(end_angle)

    def radius(self, radius: Accessor[T, float] | Number) -> AreaRadial:
        """
        Sets radius accessor function

        Parameters
        ----------
        radius : Accessor[T, float] | Number
            radius accessor function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().y(radius)

    def inner_radius(self, inner_radius: Accessor[T, float] | Number) -> AreaRadial:
        """
        Sets inner radius accessor function

        Parameters
        ----------
        inner_radius : Accessor[T, float] | Number
            radius accessor function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().y0(inner_radius)

    def outer_radius(self, outer_radius: Accessor[T, float] | Number) -> AreaRadial:
        """
        Sets outer radius accessor function

        Parameters
        ----------
        outer_radius : Accessor[T, float] | Number
            radius accessor function

        Returns
        -------
        AreaRadial
            Itself
        """
        return super().y1(outer_radius)

    def line_start_angle(self) -> LineRadial:
        """
        Returns a new radial line generator that has this radial area
        generator's current defined accessor, curve and context. The line's
        angle accessor is this area's start angle accessor, and the line's
        radius accessor is this area's inner radius accessor.


        Returns
        -------
        LineRadial
            Radial line generator based on the area parameters
        """
        line_x0 = super().line_x0()
        return (
            LineRadial()
            .angle(line_x0.get_x())
            .radius(line_x0.get_y())
            .set_defined(line_x0.get_defined())
            .set_context(line_x0.get_context())
            .set_curve(line_x0.get_curve())
        )

    def line_end_angle(self) -> LineRadial:
        """
        Returns a new radial line generator that has this radial area
        generator's current defined accessor, curve and context. The line's
        angle accessor is this area's end angle accessor, and the line's radius
        accessor is this area's inner radius accessor.

        Returns
        -------
        LineRadial
            Radial line generator based on the area parameters
        """
        line_x1 = super().line_x1()
        return (
            LineRadial()
            .angle(line_x1.get_x())
            .radius(line_x1.get_y())
            .set_defined(line_x1.get_defined())
            .set_context(line_x1.get_context())
            .set_curve(line_x1.get_curve())
        )

    def line_inner_radius(self) -> LineRadial:
        """
        Returns a new radial line generator that has this radial area
        generator's current defined accessor, curve and context. The line's
        angle accessor is this area's start angle accessor, and the line's
        radius accessor is this area's inner radius accessor.

        Returns
        -------
        LineRadial
            Radial line generator based on the area parameters
        """
        line_y0 = super().line_y0()
        return (
            LineRadial()
            .angle(line_y0.get_x())
            .radius(line_y0.get_y())
            .set_defined(line_y0.get_defined())
            .set_context(line_y0.get_context())
            .set_curve(line_y0.get_curve())
        )

    def line_outer_radius(self) -> LineRadial:
        """
        Returns a new radial line generator that has this radial area
        generator's current defined accessor, curve and context. The line's
        angle accessor is this area's start angle accessor, and the line's
        radius accessor is this area's outer radius accessor.

        Returns
        -------
        LineRadial
            Radial line generator based on the area parameters
        """
        line_y1 = super().line_y1()
        return (
            LineRadial()
            .angle(line_y1.get_x())
            .radius(line_y1.get_y())
            .set_defined(line_y1.get_defined())
            .set_context(line_y1.get_context())
            .set_curve(line_y1.get_curve())
        )

    def get_angle(self) -> Accessor[T, float]:
        return super().get_x()

    def get_start_angle(self) -> Accessor[T, float]:
        return super().get_x0()

    def get_end_angle(self) -> Accessor[T, float]:
        return super().get_x1()

    def get_radius(self) -> Accessor[T, float]:
        return super().get_y()

    def get_inner_radius(self) -> Accessor[T, float]:
        return super().get_y0()

    def get_outer_radius(self) -> Accessor[T, float]:
        return super().get_y1()


def area_radial() -> AreaRadial:
    """
    A radial area generator is like the Cartesian area generator except the x
    and y accessors are replaced with angle and radius accessors. Radial areas
    are positioned relative to the origin; use a transform to change the
    origin.

    Returns
    -------
    AreaRadial
        Radial area generator
    """
    return AreaRadial()



from __future__ import annotations

from collections.abc import Callable
from copy import deepcopy
from math import floor, inf, isnan, nan, sqrt
from typing import Generic

from ..array import argpass
from ..types import Accessor, Point2D, T
from .add import add
from .quad import Quad


def default_x(d: Point2D) -> float:
    return d[0]


def default_y(d: Point2D) -> float:
    return d[1]


class Quadtree(Generic[T]):
    """
    A quadtree recursively partitions two-dimensional space into squares,
    dividing each square into four equally-sized squares. Each distinct point
    exists in a unique leaf node; coincident points are represented by a linked
    list. Quadtrees can accelerate various spatial operations, such as the
    `Barnes–Hut approximation
    <https://en.wikipedia.org/wiki/Barnes%E2%80%93Hut_simulation>`_ for
    computing many-body forces, collision detection, and searching for nearby
    points.

    Parameters
    ----------
    x : Accessor[T, float]
        X accessor
    y : Accessor[T, float]
        Y accessor
    x0 : float
        Inclusive lower bound :math:`x_0`
    y0 : float
        Inclusive lower bound :math:`y_0`
    x1 : float
        Inclusive upper bound :math:`x_1`
    y1 : float
        Inclusive upper bound :math:`y_1`
    """

    def __init__(
        self,
        x: Accessor[T, float],
        y: Accessor[T, float],
        x0: float,
        y0: float,
        x1: float,
        y1: float,
    ):
        self._x = argpass(x)
        self._y = argpass(y)
        self._x0 = x0
        self._y0 = y0
        self._x1 = x1
        self._y1 = y1
        self._root = None

    def x(self, x: Accessor[T, float]) -> Quadtree:
        """
        Sets the x-coordinate accessor and returns the quadtree.
        The x accessor is used to derive the x coordinate of data when adding
        to and removing from the tree. It is also used when finding to
        re-access the coordinates of data previously added to the tree;
        therefore, the x and y accessors must be consistent, returning the same
        value given the same input.

        Parameters
        ----------
        x : Accessor[T, float]
            X accessor

        Returns
        -------
        Quadtree
            Itself
        """
        self._x = argpass(x)
        return self

    def y(self, y: Accessor[T, float]) -> Quadtree:
        """
        Sets the y-coordinate accessor and returns the quadtree.
        The y accessor is used to derive the y coordinate of data when adding
        to and removing from the tree. It is also used when finding to
        re-access the coordinates of data previously added to the tree;
        therefore, the x and y accessors must be consistent, returning the same
        value given the same input.

        Parameters
        ----------
        y : Accessor[T, float]
            Y accessor

        Returns
        -------
        Quadtree
            Itself
        """
        self._y = argpass(y)
        return self

    def add(self, d: T) -> Quadtree:
        """
        Adds the specified datum to the quadtree, deriving its coordinates
        :math:`(x, y)` using the current x and y accessors, and returns the
        quadtree.

        If the new point is outside the current extent of the quadtree, the
        quadtree is automatically expanded to cover the new point.

        Parameters
        ----------
        d : T
            Datum

        Returns
        -------
        Quadtree
            Itself
        """
        x = self._x(d)
        y = self._y(d)
        return add(self.cover(x, y), x, y, d)

    def add_all(self, data: list[T]) -> Quadtree:
        """
        Adds the specified iterable of data to the quadtree, deriving each
        element’s coordinates :math:`(x, y)` using the current x and y
        accessors, and return this quadtree.


        Parameters
        ----------
        data : list[T]
            List of data values

        Returns
        -------
        Quadtree
            Itself
        """
        data = list(data)
        n = len(data)
        xz = [None] * n
        yz = [None] * n
        x0 = inf
        y0 = x0
        x1 = -x0
        y1 = x1

        for i in range(n):
            d = data[i]
            x = self._x(d)
            y = self._y(d)
            if isnan(x) or isnan(y):
                continue
            xz[i] = x
            yz[i] = y
            if x < x0:
                x0 = x
            if x > x1:
                x1 = x
            if y < y0:
                y0 = y
            if y > y1:
                y1 = y

        if x0 > x1 or y0 > y1:
            return self

        self.cover(x0, y0).cover(x1, y1)

        for i in range(n):
            add(self, xz[i], yz[i], data[i])

        return self

    def cover(self, x: float, y: float) -> Quadtree:
        """
        Expands the quadtree to cover the specified point :math:`(x, y)`, and
        returns the quadtree.

        Parameters
        ----------
        x : float
            X value
        y : float
            Y value

        Returns
        -------
        Quadtree
            Itself
        """
        if isnan(x) or isnan(y):
            return self

        x0 = self._x0
        y0 = self._y0
        x1 = self._x1
        y1 = self._y1

        if isnan(x0):
            x0 = floor(x)
            y0 = floor(y)
            x1 = x0 + 1
            y1 = y0 + 1
        else:
            z = (x1 - x0) or 1
            node = self._root

            while x0 > x or x >= x1 or y0 > y or y >= y1:
                i = (y < y0) << 1 | (x < x0)
                parent = [None] * 4
                parent[i] = node
                node = parent
                z *= 2
                match i:
                    case 0:
                        x1 = x0 + z
                        y1 = y0 + z
                    case 1:
                        x0 = x1 - z
                        y1 = y0 + z
                    case 2:
                        x1 = x0 + z
                        y0 = y1 - z
                    case 3:
                        x0 = x1 - z
                        y0 = y1 - z

            if self._root and isinstance(self._root, list):
                self._root = node

        self._x0 = x0
        self._y0 = y0
        self._x1 = x1
        self._y1 = y1
        return self

    def set_extent(
        self, extent: list[tuple[float, float], tuple[float, float]]
    ) -> Quadtree:
        """
        Expands the quadtree to cover the specified points :code:`[[x0, y0],
        [x1, y1]]` and returns the quadtree.

        Parameters
        ----------
        extent : list[tuple[float, float], tuple[float, float]]
            Extent values

        Returns
        -------
        Quadtree
            Itself
        """
        return self.cover(extent[0][0], extent[0][1]).cover(extent[1][0], extent[1][1])

    def data(self) -> list[T]:
        """
        Returns an array of all data in the quadtree.

        Returns
        -------
        list[T]
            List of all data
        """
        data = []

        def visit(node):
            if isinstance(node, list):
                return
            while True:
                data.append(node["data"])
                node = node.get("next")
                if node is None:
                    break

        self.visit(visit)
        return data

    def find(self, x: float, y: float, radius: float | None = None) -> Quadtree:
        """
        Returns the datum closest to the position :math:`(x, y)` with the given
        search radius. If radius is not specified, it defaults to infinity.

        Parameters
        ----------
        x : float
            x value
        y : float
            y value
        radius : float | None
            radius value

        Returns
        -------
        Quadtree
            Itself
        """
        x0 = self._x0
        y0 = self._y0
        x3 = self._x1
        y3 = self._y1
        quads = []
        data = None
        node = self._root

        if node:
            quads.append(Quad(node, x0, y0, x3, y3))
        if radius is None:
            radius = inf
        else:
            x0 = x - radius
            y0 = y - radius
            x3 = x + radius
            y3 = y + radius
            radius *= radius

        while len(quads):
            q = quads.pop()
            node = q.node
            x1 = q.x0
            y1 = q.y0
            x2 = q.x1
            y2 = q.y1
            if not node or x1 > x3 or y1 > y3 or x2 < x0 or y2 < y0:
                continue

            if isinstance(node, list):
                xm = (x1 + x2) * 0.5
                ym = (y1 + y2) * 0.5

                quads.extend(
                    (
                        Quad(node[3], xm, ym, x2, y2),
                        Quad(node[2], x1, ym, xm, y2),
                        Quad(node[1], xm, y1, x2, ym),
                        Quad(node[0], x1, y1, xm, ym),
                    )
                )

                i = (y >= ym) << 1 | (x >= xm)
                if i:
                    q = quads[-1]
                    quads[-1] = quads[-1 - i]
                    quads[-1 - i] = q
            else:
                dx = x - self._x(node["data"])
                dy = y - self._y(node["data"])
                d2 = dx * dx + dy * dy
                if d2 < radius:
                    radius = d2
                    d = sqrt(d2)
                    x0 = x - d
                    y0 = y - d
                    x3 = x + d
                    y3 = y + d
                    data = node["data"]

        return data

    def remove(self, d: T) -> Quadtree:
        """
        Removes the specified datum from the quadtree, deriving its coordinates
        :math:`(x,y)` using the current x and y accessors, and returns the quadtree.

        Parameters
        ----------
        d : T
            Datum

        Returns
        -------
        Quadtree
            Itself
        """
        x = self._x(d)
        y = self._y(d)
        if isnan(x) or isnan(y):
            return self

        parent = None
        previous = None
        retainer = None
        node = self._root
        x0 = self._x0
        y0 = self._y0
        x1 = self._x1
        y1 = self._y1

        if node is None:
            return self

        if isinstance(node, list):
            while True:
                xm = (x0 + x1) * 0.5
                right = x >= xm
                if right:
                    x0 = xm
                else:
                    x1 = xm

                ym = (y0 + y1) * 0.5
                bottom = y >= ym
                if bottom:
                    y0 = ym
                else:
                    y1 = ym

                parent = node
                i = bottom << 1 | right
                node = node[i]
                if node is None:
                    return self
                if isinstance(node, dict):
                    break
                if parent[(i + 1) & 3] or parent[(i + 2) & 3] or parent[(i + 3) & 3]:
                    retainer = parent
                    j = i

        while node["data"] is not d:
            previous = node
            node = node.get("next")
            if node is None:
                return self
        next = node.pop("next", None)

        if previous:
            if next:
                previous["next"] = next
            else:
                previous.pop("next")
            return self

        if parent is None:
            self._root = next
            return self

        parent[i] = next

        node = parent[0] or parent[1] or parent[2] or parent[3]
        if node == (parent[3] or parent[2] or parent[1] or parent[0]) and isinstance(
            node, dict
        ):
            if retainer:
                retainer[j] = node
            else:
                self._root = node

        return self

    def remove_all(self, data: list[T]) -> Quadtree:
        """
        Removes the specified data from the quadtree, deriving their
        coordinates :math:`(x,y)` using the current x and y accessors, and
        returns the quadtree.

        Parameters
        ----------
        data : list[T]
            List of data

        Returns
        -------
        Quadtree
            Itself
        """
        for d in data:
            self.remove(d)
        return self

    def visit(
        self, callback: Callable[[list | dict | None, float, float, float, float], bool]
    ) -> Quadtree:
        """
        Visits each node in the quadtree in pre-order traversal, invoking the
        specified callback with arguments :code:`node`, :code:`x0`, :code:`y0`,
        :code:`x1`, :code:`y1` for each node, where node is the node being
        visited, :math:`(x_0, y_0)` are the lower bounds of the node, and
        :math:`(x_1, y_1)` are the upper bounds, and returns the quadtree.
        (Assuming that positive x is right and positive y is down, as is
        typically the case in Canvas and SVG, :math:`(x_0, y_0)` is the
        top-left corner and :math:`(x_1, y_1)` is the lower-right corner;
        however, the coordinate system is arbitrary, so more formally
        :math:`x_0 \\le x_1` and :math:`y0 \\le y1`.)

        If the callback returns true for a given node, then the children of
        that node are not visited; otherwise, all child nodes are visited. This
        can be used to quickly visit only parts of the tree, for example when
        using the `Barnes–Hut approximation
        <https://en.wikipedia.org/wiki/Barnes%E2%80%93Hut_simulation>`_. Note,
        however, that child quadrants are always visited in sibling order:
        top-left, top-right, bottom-left, bottom-right. In cases such as
        search, visiting siblings in a specific order may be faster.

        Parameters
        ----------
        callback : Callable[[list | dict | None, float, float, float, float], bool]
            Callback function

        Returns
        -------
        Quadtree
            Itself

        Notes
        -----
        The number of arguments for the :code:`callback` function is
        automatically determined. Therefore, you can pass 0 to 5 arguments to
        the :code:`callback` function. For example, `def callback(node):` is a
        valid function.
        """
        callback = argpass(callback)
        quads = []
        node = self._root
        if node:
            quads.append(Quad(node, self._x0, self._y0, self._x1, self._y1))
        while quads:
            q = quads.pop()
            node = q.node
            x0 = q.x0
            y0 = q.y0
            x1 = q.x1
            y1 = q.y1
            if not callback(node, x0, y0, x1, y1) and isinstance(node, list):
                xm = (x0 + x1) * 0.5
                ym = (y0 + y1) * 0.5
                child = node[3]
                if child:
                    quads.append(Quad(child, xm, ym, x1, y1))
                child = node[2]
                if child:
                    quads.append(Quad(child, x0, ym, xm, y1))
                child = node[1]
                if child:
                    quads.append(Quad(child, xm, y0, x1, ym))
                child = node[0]
                if child:
                    quads.append(Quad(child, x0, y0, xm, ym))
        return self

    def visit_after(
        self, callback: Callable[[dict, float, float, float, float], bool]
    ) -> Quadtree:
        """
        Visits each node in the quadtree in post-order traversal, invoking the
        specified callback with arguments :code:`node`, :code:`x0`, :code:`y0`,
        :code:`x1`, :code:`y1` for each node, where node is the node being
        visited, :code:`(x_0, y_0)` are the lower bounds of the node, and
        :code:`(x_1, y_1)` are the upper bounds, and returns the quadtree.
        (Assuming that positive :code:`x` is right and positive :code:`y` is
        down, as is typically the case in Canvas and SVG, :code:`(x_0, y_0)` is
        the top-left corner and :code:`(x_1, y_1)` is the lower-right corner;
        however, the coordinate system is arbitrary, so more formally
        :math:`x_0 \\lt x_1` and :math:`y0 \\lt y1`.). Returns root.

        Parameters
        ----------
        callback : Callable[[dict, float, float, float, float], bool]
            Callback function

        Returns
        -------
        Quadtree
            Itself

        Notes
        -----
        The number of arguments for the :code:`callback` function is
        automatically determined. Therefore, you can pass 0 to 5 arguments to
        the :code:`callback` function. For example, `def callback(node):` is a
        valid function.
        """
        callback = argpass(callback)
        quads = []
        next = []
        if self._root:
            quads.append(Quad(self._root, self._x0, self._y0, self._x1, self._y1))
        while quads:
            q = quads.pop()
            node = q.node
            if isinstance(node, list):
                x0 = q.x0
                y0 = q.y0
                x1 = q.x1
                y1 = q.y1
                xm = (x0 + x1) * 0.5
                ym = (y0 + y1) * 0.5
                child = node[0]
                if child:
                    quads.append(Quad(child, x0, y0, xm, ym))
                child = node[1]
                if child:
                    quads.append(Quad(child, xm, y0, x1, ym))
                child = node[2]
                if child:
                    quads.append(Quad(child, x0, ym, xm, y1))
                child = node[3]
                if child:
                    quads.append(Quad(child, xm, ym, x1, y1))
            next.append(q)
        while len(next) != 0:
            q = next.pop()
            callback(q.node, q.x0, q.y0, q.x1, q.y1)
        return self

    def copy(self) -> Quadtree:
        """
        Returns a deep copy of itself.

        Returns
        -------
        Quadtree
            Deep copy of itself
        """
        return deepcopy(self)

    def size(self) -> float:
        """
        Returns the total number of data in the quadtree.

        Returns
        -------
        float
            Total number of data in sthe quadtree.
        """
        size = [0]

        def visit(node: list | dict | None):
            if isinstance(node, list):
                return
            while True:
                size[0] += 1
                node = node.get("next")
                if node is None:
                    break

        self.visit(visit)
        return size[0]

    def get_extent(self) -> list[tuple[float, float], tuple[float, float]]:
        if isnan(self._x0):
            return None
        return [[self._x0, self._y0], [self._x1, self._y1]]

    def get_root(self) -> list | dict | None:
        return self._root

    def get_x(self) -> Accessor[T, float]:
        return self._x

    def get_y(self) -> Accessor[T, float]:
        return self._y


def quadtree(
    data: list[T] | None = None,
    x: Accessor[T, float] = None,
    y: Accessor[T, float] = None,
) -> Quadtree[T]:
    """
    Creates a new, empty quadtree with an empty extent and the default x and y
    accessors. If data is specified, adds the specified iterable of data to the
    quadtree.

    Parameters
    ----------
    data : list[T] | None
        List of data values
    x : Accessor[T, float]
        X accessor
    y : Accessor[T, float]
        Y accessor

    Returns
    -------
    Quadtree[T]
        Quadtree object
    """
    tree = Quadtree(
        default_x if x is None else x, default_y if y is None else y, nan, nan, nan, nan
    )
    return tree if data is None else tree.add_all(data)



from __future__ import annotations

from collections.abc import Callable

from ..color import rgb as color_rgb
from .basis import interpolate_basis
from .basis_closed import interpolate_basis_closed
from .color import color as nogamma
from .color import gamma


class RGBGammaInterpolator:
    def __init__(self, y):
        self.color = gamma(y)

    def __call__(self, start: str, end: str) -> Callable[[float], str]:
        """
        Returns an RGB color space interpolator between the two colors a and b
        with a configurable gamma. If the gamma is not specified, it defaults to
        1.0. The colors a and b need not be in RGB; they will be converted to RGB
        using :code:`d3.rgb`. The return value of the interpolator is an RGB string.

        Parameters
        ----------
        start : str
            Color string a
        end : str
            Color string b

        Returns
        -------
        Callable[[float], str]
            Interpolator
        """
        start = color_rgb(start)
        end = color_rgb(end)
        r = self.color(start.r, end.r)
        g = self.color(start.g, end.g)
        b = self.color(start.b, end.b)
        opacity = nogamma(start.opacity, end.opacity)

        def interpolate(t):
            start.r = r(t)
            start.g = g(t)
            start.b = b(t)
            start.opacity = opacity(t)
            return str(start)

        return interpolate

    def set_gamma(self, y: float) -> RGBGammaInterpolator:
        """
        Sets gamma and returns itself. See `gamma error in picture scaling
        <http://www.ericbrasseur.org/gamma.html?i=1>`_ for more information.

        Parameters
        ----------
        y : float
            Gamma value

        Returns
        -------
        RGBGammaInterpolator
            Itself
        """
        self.color = gamma(y)
        return self


interpolate_rgb = RGBGammaInterpolator(1)
interpolate_rgb.__doc__ = """
Returns an interpolator between the two colors start and end.

Parameters
----------
start : Number
    start value
end : Number
    end value

Returns
-------
Callable[[float], str]
    Interpolator function

Examples
--------

>>> interpolator = d3.interpolate_rgb("red", "blue")
>>> interpolator(0)
'rgb(255, 0, 0)'
>>> interpolator(0.5)
'rgb(128, 0, 128)'
>>> interpolator(1)
'rgb(0, 0, 255)'
"""


class RGBSplineInterpolator:
    def __init__(self, spline):
        self.spline = spline

    def __call__(self, colors):
        n = len(colors)
        r = [0] * n
        g = [0] * n
        b = [0] * n

        for i, color in enumerate(colors):
            color = color_rgb(color)
            r[i] = color.r or 0
            g[i] = color.g or 0
            b[i] = color.b or 0

        r_spline = self.spline(r)
        g_spline = self.spline(g)
        b_spline = self.spline(b)

        color = color_rgb(0, 0, 0, 1)

        def interpolate(t):
            color.r = r_spline(t)
            color.g = g_spline(t)
            color.b = b_spline(t)
            return str(color)

        return interpolate


def interpolate_rgb_basis(colors: list[str]) -> Callable[[float], str]:
    """
    Returns a uniform nonrational B-spline interpolator through the specified
    array of colors, which are converted to RGB color space. Implicit control
    points are generated such that the interpolator returns colors[0] at t = 0
    and colors[-1] at t = 1. Opacity interpolation is not currently supported.
    See also d3.interpolate_basis, and see scale_chromatic for examples.

    Parameters
    ----------
    colors : list[str]
         List of colors to interpolate

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_rgb_basis(["red", "green", "blue"])
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(0.5)
    'rgb(42, 85, 42)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    """
    return RGBSplineInterpolator(interpolate_basis)(colors)


def interpolate_rgb_basis_closed(colors: list[str]) -> Callable[[float], str]:
    """
    Returns a uniform nonrational B-spline interpolator through the specified
    array of colors, which are converted to RGB color space. The control points
    are implicitly repeated such that the resulting spline has cyclical :math:`C^2`
    continuity when repeated around t in [0,1]; this is useful, for example, to
    create cyclical color scales. Opacity interpolation is not currently supported.
    See also d3.interpolate_basis_closed, and see scale_chromatic for examples.

    Parameters
    ----------
    colors : list[str]
         List of colors to interpolate

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_rgb_basis_closed(["red", "green", "blue"])
    >>> interpolator(0)
    'rgb(170, 21, 42)'
    >>> interpolator(0.5)
    'rgb(11, 61, 122)'
    >>> interpolator(1)
    'rgb(255, 0, 255)'
    """
    return RGBSplineInterpolator(interpolate_basis_closed)(colors)



from collections.abc import Callable

from ..types import Number


def interpolate_number(a: Number, b: Number) -> Callable[[float], float]:
    """
    Returns an interpolator between the two numbers a and b.

    Parameters
    ----------
    a : Number
        a value
    b : Number
        b value

    Returns
    -------
    Callable[[float], float]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_number(10, 20)
    >>> interpolator(0)
    10.0
    >>> interpolator(1)
    20.0
    >>> interpolator(0.5)
    15.0
    """
    a, b = float(a), float(b)
    return lambda t: a * (1 - t) + b * t



from copy import copy, deepcopy
from typing import Generic

from lxml import etree

from ..types import T


class EnterNode(Generic[T]):
    """
    Enter node which holds parent element, a datum and next element in some
    specific cases.

    Parameters
    ----------
    parent : etree.Element
        Parent node
    datum : T
        Data to hold
    """

    def __init__(self, parent: etree.Element, datum: T):
        self._next = None
        self._parent = parent
        self.__data__ = datum

    def __str__(self):
        if self._parent is None:
            return f"EnterNode({self._parent}, {self.__data__})"
        tag = self._parent.tag
        class_name = self._parent.attrib.get("class")
        if class_name:
            return f"EnterNode({tag}.{class_name}, {self.__data__})"
        return f"EnterNode({tag}, {self.__data__})"

    def clone(self, deep: bool = False):
        copy_func = deepcopy if deep else copy
        node = EnterNode(
            copy_func(self._parent),
            copy_func(self.__data__),
        )
        node._next = copy_func(self._next)
        return node

    def __repr__(self):
        return str(self)



from ...types import Point2D
from ..common import Projection
from .projection import geo_projection

EPSILON = 1e-6


class NaturalEarth1Raw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        phi2 = phi * phi
        phi4 = phi2 * phi2
        return [
            lambda_
            * (
                0.8707
                - 0.131979 * phi2
                + phi4 * (-0.013791 + phi4 * (0.003971 * phi2 - 0.001529 * phi4))
            ),
            phi
            * (
                1.007226
                + phi2
                * (0.015085 + phi4 * (-0.044475 + 0.028874 * phi2 - 0.005916 * phi4))
            ),
        ]

    def invert(self, x: float, y: float) -> Point2D:
        phi = y
        i = 25

        while True:
            phi2 = phi * phi
            phi4 = phi2 * phi2
            delta = (
                phi
                * (
                    1.007226
                    + phi2
                    * (
                        0.015085
                        + phi4 * (-0.044475 + 0.028874 * phi2 - 0.005916 * phi4)
                    )
                )
                - y
            ) / (
                1.007226
                + phi2
                * (
                    0.015085 * 3
                    + phi4
                    * (-0.044475 * 7 + 0.028874 * 9 * phi2 - 0.005916 * 11 * phi4)
                )
            )
            phi -= delta
            i -= 1
            if abs(delta) <= EPSILON or i <= 0:
                break

        phi2 = phi * phi
        return [
            x
            / (
                0.8707
                + phi2
                * (
                    -0.131979
                    + phi2
                    * (-0.013791 + phi2 * phi2 * phi2 * (0.003971 - 0.001529 * phi2))
                )
            ),
            phi,
        ]


def geo_natural_earth_1() -> Projection:
    """
    The Natural Earth projection is a pseudocylindrical projection designed by
    Tom Patterson. It is neither conformal nor equal-area, but appealing to the
    eye for small-scale maps of the whole world.

    Returns
    -------
    Projection
        Projection object
    """
    return geo_projection(NaturalEarth1Raw()).scale(175.295)



from ...color import cubehelix
from ...interpolate import interpolate_cubehelix_long

interpolate_cubehelix_default = interpolate_cubehelix_long(
    cubehelix(300, 0.5, 0.0), cubehelix(-240, 0.5, 1.0)
)



from math import isnan

from ...types import T
from ..series import Series


def offset_none(series: list[Series[T]], order: list[int]):
    """
    Applies a zero baseline.

    Parameters
    ----------
    series : list[Series[T]]
        List of series
    order : list[int]
        List of ordered indices
    """
    n = len(series)
    if n == 0:
        return
    s1: Series[T] = series[order[0]]
    m = len(s1)
    for i in range(1, n):
        s0: Series[T] = s1
        s1: Series[T] = series[order[i]]
        for j in range(m):
            s1[j][0] = s0[j][0] if isnan(s0[j][1]) else s0[j][1]
            s1[j][1] += s1[j][0]



from abc import ABC, abstractclassmethod
from math import cos, inf, isnan, pi, sin

from ...types import Number


def isvaluable(x: Number | None) -> bool:
    return x is not None and not isnan(x) and x


def point_radial(x: Number, y: Number) -> tuple[Number, Number]:
    x -= pi * 0.5
    return y * cos(x), y * sin(x)


def sign(x: Number) -> bool:
    return -1 if x < 0 else 1


def fdiv(y: float, x: float) -> float:
    try:
        return y / x
    except ZeroDivisionError:
        return inf if y > 0 else -inf


class Curve(ABC):
    """
    Curves are typically not used directly, instead being passed to line.curve
    and area.curve. However, you can define your own curve implementation
    should none of the built-in curves satisfy your needs using the following
    interface; see the curveLinear source for an example implementation. You
    can also use this low-level interface with a built-in curve type as an
    alternative to the line and area generators.
    """

    @abstractclassmethod
    def area_start(self):
        """
        Indicates the start of a new area segment. Each area segment consists
        of exactly two line segments: the topline, followed by the baseline,
        with the baseline points in reverse order.

        """
        ...

    @abstractclassmethod
    def area_end(self):
        """
        Indicates the end of the current area segment.
        """
        ...

    @abstractclassmethod
    def line_start(self):
        """
        Indicates the start of a new line segment. Zero or more points will
        follow.
        """
        ...

    @abstractclassmethod
    def line_end(self):
        """
        Indicates the end of the current line segment.
        """
        ...

    @abstractclassmethod
    def point(self, x: float, y: float):
        """
        Indicates a new point in the current line segment with the given x- and y-values.

        Parameters
        ----------
        x : float
            x value
        y : float
            y value
        """
        ...



from math import sqrt


def symbol_diamond2(context, size):
    r = sqrt(size) * 0.62625
    context.move_to(0, -r)
    context.line_to(r, 0)
    context.line_to(0, r)
    context.line_to(-r, 0)
    context.close_path()



from .area import geo_area
from .bounds import geo_bounds
from .centroid import geo_centroid
from .circle import geo_circle
from .clip import (
    geo_clip_antimeridian,
    geo_clip_circle,
    geo_clip_rectangle,
)
from .contains import geo_contains
from .distance import geo_distance
from .graticule import (
    geo_graticule,
    geo_graticule_10,
)
from .interpolate import geo_interpolate
from .length import geo_length
from .path import GeoPath as geo_path
from .projection import (
    geo_albers,
    geo_albers_usa,
    geo_azimuthal_equal_area,
    geo_azimuthal_equidistant,
    geo_conic_conformal,
    geo_conic_equal_area,
    geo_conic_equidistant,
    geo_equal_earth,
    geo_equirectangular,
    geo_gnomonic,
    geo_identity,
    geo_mercator,
    geo_natural_earth_1,
    geo_orthographic,
    geo_projection,
    geo_projection_mutator,
    geo_stereographic,
    geo_transverse_mercator,
)
from .rotation import geo_rotation
from .stream import geo_stream
from .transform import geo_transform

__all__ = [
    "geo_albers",
    "geo_albers_usa",
    "geo_area",
    "geo_azimuthal_equal_area",
    "geo_azimuthal_equidistant",
    "geo_bounds",
    "geo_centroid",
    "geo_circle",
    "geo_clip_antimeridian",
    "geo_clip_circle",
    "geo_clip_rectangle",
    "geo_conic_conformal",
    "geo_conic_equal_area",
    "geo_conic_equidistant",
    "geo_contains",
    "geo_distance",
    "geo_equal_earth",
    "geo_equirectangular",
    "geo_gnomonic",
    "geo_graticule",
    "geo_graticule_10",
    "geo_identity",
    "geo_interpolate",
    "geo_length",
    "geo_mercator",
    "geo_natural_earth_1",
    "geo_orthographic",
    "geo_path",
    "geo_projection",
    "geo_projection_mutator",
    "geo_rotation",
    "geo_stereographic",
    "geo_stream",
    "geo_transform",
    "geo_transverse_mercator",
]



from .decompose import decompose, identity


def parseCss(value):
    try:
        from cssselect import parser

        m = parser.parse_style_attribute(value)[0]
        return decompose(m.a, m.b, m.c, m.d, m.e, m.f)
    except ImportError:
        print("cssselect library not found. Install it for CSS parsing support.")
        return identity


def parseSvg(value):
    if value is None:
        return identity
    try:
        from svg.path import parse_path

        path = parse_path(value)
        m = path.transform()
        return decompose(m.a, m.b, m.c, m.d, m.e, m.f)
    except ImportError:
        print("svg.path library not found. Install it for SVG parsing support.")
        return identity



from __future__ import annotations

import math
from bisect import bisect
from collections.abc import Callable, Iterable
from typing import Any

from .argpass import argpass
from .extent import extent
from .nice import nice
from .threshold import threshold_sturges
from .ticks import tick_increment, ticks


def identity(x, *args):
    return x


def constant(*obj):
    if len(obj) == 1:
        obj = obj[0]

    def wrapper(*args):
        return obj

    return wrapper


class Bin:
    """
    Bin quantitative values into consecutive, non-overlapping
    intervals, as in histograms
    """

    def __init__(self):
        self._list = []
        self.x0 = None
        self.x1 = None

    def __len__(self) -> int:
        return len(self._list)

    def __getitem__(self, index: int):
        return self._list[index]

    def __setitem__(self, index, item):
        self._list[index] = item

    def append(self, item: int | float):
        self._list.append(item)

    def __str__(self):
        return f"Bin({self._list}, x0={self.x0}, x1={self.x1})"

    def __repr__(self):
        return str(self)

    def __eq__(self, bin) -> bool:
        return self._list == bin._list and self.x0 == self.x0 and self.x1 == self.x1


class bin:
    """
    Bin generator with the default settings
    """

    def __init__(self):
        self._value = argpass(identity)
        self._domain = extent
        self._threshold = threshold_sturges

    def __call__(self, data: Iterable[int | float]) -> list[Bin]:
        """
        Returns an array of bins, where each bin is an array
        containing the associated elements from the input data.

        Parameters
        ----------
        data : Iterable[int | float]
            Data samples

        Returns
        -------
        list[Bin]
            Array of bins

        Examples
        --------

        >>> d3.bin()([0, 0, 0, 10, 20, 20])
        [Bin([0, 0, 0], x0=0, x1=5), Bin([], x0=5, x1=10), Bin([10], x0=10, x1=15), Bin([], x0=15, x1=20), Bin([20, 20], x0=20, x1=25)]
        """
        if not isinstance(data, list):
            data = list(data)

        n = len(data)
        step = math.nan
        values = [None] * n

        for i in range(n):
            values[i] = self._value(data[i], i, data)

        xz = self._domain(values)
        x0, x1 = xz[0], xz[1]
        tz = self._threshold(values, x0, x1)

        if not isinstance(tz, (list, tuple)):
            max_ = x1
            tn = int(tz)
            if self._domain == extent:
                x0, x1 = nice(x0, x1, tn)
            tz = ticks(x0, x1, tn)

            if tz[0] <= x0:
                step = tick_increment(x0, x1, tn)

            if tz[-1] >= x1:
                if max_ >= x1 and self._domain == extent:
                    step = tick_increment(x0, x1, tn)
                    if math.isfinite(step):
                        if step > 0:
                            x1 = (math.floor(x1 / step) + 1) * step
                        elif step < 0:
                            x1 = (math.ceil(x1 * -step) + 1) / -step
                else:
                    tz.pop()

        m = len(tz)
        a = 0
        b = m
        while a < len(tz) and tz[a] <= x0:
            a += 1
        while b >= 0 and tz[b - 1] > x1:
            b -= 1
        if a or b < m:
            tz = tz[a:b]
            m = b - a

        bins = [None] * (m + 1)

        for i in range(m + 1):
            bins[i] = Bin()
            bins[i].x0 = tz[i - 1] if i > 0 else x0
            bins[i].x1 = tz[i] if i < m else x1

        if math.isfinite(step):
            if step > 0:
                for i in range(n):
                    if (x := values[i]) is not None and x0 <= x <= x1:
                        bins[min(m, math.floor((x - x0) / step))].append(data[i])
            elif step < 0:
                for i in range(n):
                    if (x := values[i]) is not None and x0 <= x <= x1:
                        j = math.floor((x0 - x) * step)
                        if j < len(tz):
                            bins[min(m, j + (tz[j] <= x))].append(data[i])
                        else:
                            bins[m].append(data[i])
        else:
            for i in range(n):
                if (x := values[i]) is not None and x0 <= x <= x1:
                    bins[bisect(tz, x, 0, m)].append(data[i])

        return bins

    def set_value(
        self, value: Callable[[int | float, int, Iterable[int | float]], float] | Any
    ) -> bin:
        """
        Sets value

        Parameters
        ----------
        Value : Callable[[int | float, int, Iterable[int | float]], float] | Any
            Value function or constant object

        Returns
        -------
        bin
            Itself
        """
        self._value = argpass(value if callable(value) else constant(value))
        return self

    def set_domain(self, domain: Callable[[float], float] | tuple[float, float]) -> bin:
        """
        Sets domain

        Parameters
        ----------
        obj : Callable[[float], float] | tuple[float, float]
            Domain function or domain tuple

        Returns
        -------
        bin
            Itself
        """
        self._domain = domain if callable(domain) else constant(domain[0], domain[1])
        return self

    def set_thresholds(
        self, thresholds: Callable[[list[float | None]], float] | Any
    ) -> bin:
        """
        Sets thresholds

        Parameters
        ----------
        thresholds : Callable[[list[float | None]], float] | Any
            Object or function

        Returns
        -------
        bin
            Itself
        """
        self._threshold = thresholds if callable(thresholds) else constant(thresholds)
        return self

    def get_value(self):
        return self._value

    def get_domain(self):
        return self._domain

    def get_thresholds(self):
        return self._threshold



import math
from collections.abc import Callable

from .color import hue


def interpolate_hue(a: float, b: float) -> Callable[[float], float]:
    """
    Returns an interpolator between the two hue angles a and b.
    If either hue is NaN, the opposing value is used. The shortest
    path between hues is used. The return value of the interpolator
    is a number in :math:`[0, 360)`.

    Parameters
    ----------
    a : float
        Hue angle a
    b : float
        Hue angle b

    Returns
    -------
    Callable[[float], float]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_hue(45, 180)
    >>> interpolator(0)
    45.0
    >>> interpolator(1)
    180.0
    >>> interpolator(0.5)
    112.5
    """
    i = hue(float(a), float(b))

    def interpolate(t):
        x = i(t)
        return x - 360 * math.floor(x / 360)

    return interpolate



from math import atan, exp, inf, log, pi, tan

from ...types import Point2D, Vec2D
from ..common import Projection, RawProjection
from ..rotation import geo_rotation
from .projection import ProjectionMutator

TAU = 2 * pi
half_pi = pi / 2


class MercatorRaw:
    def __call__(self, lambda_: float, phi: float) -> Point2D:
        t = tan((half_pi + phi) / 2)
        return [lambda_, log(t) if t > 0.0 else -inf]

    def invert(self, x: float, y: float) -> Point2D:
        return [x, 2 * atan(exp(y)) - half_pi]


class MercatorProjection(ProjectionMutator):
    def __init__(self, project: RawProjection):
        ProjectionMutator.__init__(self, project)
        self._project = project
        self._mx0 = None
        self._my0 = None
        self._mx1 = None
        self._my1 = None
        self.reclip()

    def scale(self, k: float) -> Projection:
        super().scale(k)
        return self.reclip()

    def translate(self, translation: Vec2D) -> Projection:
        super().translate(translation)
        return self.reclip()

    def set_center(self, center: Point2D) -> Projection:
        super().set_center(center)
        return self.reclip()

    def set_clip_extent(
        self, clip_extent: tuple[Point2D, Point2D] | None = None
    ) -> Projection:
        if clip_extent is None:
            self._mx0 = None
            self._my0 = None
            self._mx1 = None
            self._my1 = None
        else:
            self._mx0 = clip_extent[0][0]
            self._my0 = clip_extent[0][1]
            self._mx1 = clip_extent[1][0]
            self._my1 = clip_extent[1][1]
        return self.reclip()

    def get_clip_extent(self) -> tuple[Point2D, Point2D] | None:
        if self._mx0 is None:
            return None
        else:
            return [[self._mx0, self._my0], [self._mx1, self._my1]]

    def reclip(self) -> Projection:
        k = pi * self.get_scale()
        t = self(geo_rotation(self.get_rotation()).invert([0, 0]))
        if self._mx0 is None:
            return super().set_clip_extent([[t[0] - k, t[1] - k], [t[0] + k, t[1] + k]])
        elif isinstance(self._project, MercatorRaw):
            return super().set_clip_extent(
                [
                    [max(t[0] - k, self._mx0), self._my0],
                    [min(t[0] + k, self._mx1), self._my1],
                ]
            )
        else:
            return super().set_clip_extent(
                [
                    [self._mx0, max(t[1] - k, self._my0)],
                    [self._mx1, min(t[1] + k, self._my1)],
                ]
            )


def geo_mercator() -> Projection:
    """
    The spherical Mercator projection. Defines a default
    :func:`Projection.set_clip_extent
    <detroit.geo.common.Projection.set_clip_extent>` such that the world is
    projected to a square, clipped to approximately :math:`\\pm 85°` latitude.

    Returns
    -------
    Projection
        Projection object
    """
    return MercatorProjection(MercatorRaw()).scale(961 / TAU)



from __future__ import annotations

from collections.abc import Callable
from itertools import chain
from math import ceil
from typing import Iterable

from ..types import GeoJSON, Point2D

EPSILON = 1e-6


def frange(start: float, stop: float, step: float) -> list[float]:
    return [start + i * step for i in range(max(0, ceil((stop - start) / step)))]


def graticule_x(y0: float, y1: float, dy: float) -> Callable[[float], list[Point2D]]:
    ly = frange(y0, y1 - EPSILON, dy) + [y1]

    def local_graticule_x(x: float) -> Point2D:
        return [[x, y] for y in ly]

    return local_graticule_x


def graticule_y(x0: float, x1: float, dx: float) -> Callable[[float], list[Point2D]]:
    lx = frange(x0, x1 - EPSILON, dx) + [x1]

    def local_graticule_y(y: float) -> Point2D:
        return [[x, y] for x in lx]

    return local_graticule_y


class Graticule:
    def __init__(self):
        self._x0 = 0.0
        self._x1 = 0.0
        self._y0 = 0.0
        self._y1 = 0.0

        self._X0 = 0.0
        self._X1 = 0.0
        self._Y0 = 0.0
        self._Y1 = 0.0

        self._dx = 10
        self._dy = 10
        self._DX = 90
        self._DY = 360
        self._x = None
        self._y = None
        self._X = None
        self._Y = None
        self._precision = 2.5

        self.set_extent_major([[-180, -90 + EPSILON], [180, 90 - EPSILON]])
        self.set_extent_minor([[-180, -80 - EPSILON], [180, 80 + EPSILON]])

    def __call__(self) -> GeoJSON:
        """
        Returns a GeoJSON MultiLineString geometry object representing all
        meridians and parallels for this graticule.


        Returns
        -------
        GeoJSON
            GeoJSON object
        """
        return {"type": "MultiLineString", "coordinates": list(self._lines())}

    def lines(self) -> list[GeoJSON]:
        """
        Returns an array of GeoJSON LineString geometry objects, one for each
        meridian or parallel for this graticule.

        Returns
        -------
        list[GeoJSON]
            List of GeoJSON LineString geometry objects
        """

        def coordinates(coordinates: list[Point2D]) -> GeoJSON:
            return {"type": "LineString", "coordinates": coordinates}

        return list(map(coordinates, self._lines()))

    def _lines(self) -> Iterable[list[Point2D]]:
        def filter_x(x: float) -> bool:
            return abs(x % self._DX) > EPSILON

        def filter_y(y: float) -> bool:
            return abs(y % self._DY) > EPSILON

        return chain(
            map(
                self._X,
                frange(ceil(self._X0 / self._DX) * self._DX, self._X1, self._DX),
            ),
            map(
                self._Y,
                frange(ceil(self._Y0 / self._DY) * self._DY, self._Y1, self._DY),
            ),
            map(
                self._x,
                filter(
                    filter_x,
                    frange(ceil(self._x0 / self._dx) * self._dx, self._x1, self._dx),
                ),
            ),
            map(
                self._y,
                filter(
                    filter_y,
                    frange(ceil(self._y0 / self._dy) * self._dy, self._y1, self._dy),
                ),
            ),
        )

    def outline(self) -> GeoJSON:
        """
        Returns a GeoJSON Polygon geometry object representing the outline of
        this graticule, i.e. along the meridians and parallels defining its
        extent.

        Returns
        -------
        GeoJSON
            GeoJSON object
        """
        return {
            "type": "Polygon",
            "coordinates": [
                self._X(self._X0)
                + self._Y(self._Y1)[1:]
                + self._X(self._X1)[-2::-1]
                + self._Y(self._Y0)[-2::-1]
            ],
        }

    def set_extent(self, extent: tuple[Point2D, Point2D]) -> Graticule:
        """
        Sets the major and minor extents of this graticule.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values

        Returns
        -------
        Graticule
            Itself
        """
        return self.set_extent_major(extent).set_extent_minor(extent)

    def set_extent_major(self, extent: tuple[Point2D, Point2D]) -> Graticule:
        """
        Sets the major extent of this graticule.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values

        Returns
        -------
        Graticule
            Itself
        """
        self._X0 = extent[0][0]
        self._Y0 = extent[0][1]
        self._X1 = extent[1][0]
        self._Y1 = extent[1][1]
        if self._X0 > self._X1:
            extent = self._X0
            self._X0 = self._X1
            self._X1 = extent
        if self._Y0 > self._Y1:
            extent = self._Y0
            self._Y0 = self._Y1
            self._Y1 = extent
        return self.set_precision(self.get_precision())

    def set_extent_minor(self, extent: tuple[Point2D, Point2D]) -> Graticule:
        """
        Sets the minor extent of this graticule.

        Parameters
        ----------
        extent : tuple[Point2D, Point2D]
            Extent values

        Returns
        -------
        Graticule
            Itself
        """
        self._x0 = extent[0][0]
        self._y0 = extent[0][1]
        self._x1 = extent[1][0]
        self._y1 = extent[1][1]
        if self._x0 > self._x1:
            extent = self._x0
            self._x0 = self._x1
            self._x1 = extent
        if self._y0 > self._y1:
            extent = self._y0
            self._y0 = self._y1
            self._y1 = extent
        return self.set_precision(self.get_precision())

    def set_step(self, step: Point2D) -> Graticule:
        """
        Sets the major and minor step for this graticule.

        Parameters
        ----------
        step : Point2D
            Step values

        Returns
        -------
        Graticule
            Itself
        """
        return self.set_step_major(step).set_step_minor(step)

    def set_step_major(self, step: Point2D) -> Graticule:
        """
        Sets the major step for this graticule.

        Parameters
        ----------
        step : Point2D
            Step values

        Returns
        -------
        Graticule
            Itself
        """
        self._DX = step[0]
        self._DY = step[1]
        return self

    def set_step_minor(self, step: Point2D) -> Graticule:
        """
        Sets the minor step for this graticule.

        Parameters
        ----------
        step : Point2D
            Step values

        Returns
        -------
        Graticule
            Itself
        """
        self._dx = step[0]
        self._dy = step[1]
        return self

    def set_precision(self, precision: float) -> Graticule:
        """
        Sets the precision for this graticule.

        Parameters
        ----------
        precision : float
            Precision value

        Returns
        -------
        Graticule
            Itself
        """
        self._precision = precision
        self._x = graticule_x(self._y0, self._y1, 90)
        self._y = graticule_y(self._x0, self._x1, precision)
        self._X = graticule_x(self._Y0, self._Y1, 90)
        self._Y = graticule_y(self._X0, self._X1, precision)
        return self

    def get_extent(self) -> tuple[Point2D, Point2D]:
        return self.get_extent_minor()

    def get_extent_major(self) -> tuple[Point2D, Point2D]:
        return [[self._X0, self._Y0], [self._X1, self._Y1]]

    def get_extent_minor(self) -> tuple[Point2D, Point2D]:
        return [[self._x0, self._y0], [self._x1, self._y1]]

    def get_step(self) -> Point2D:
        return self.get_step_minor()

    def get_step_major(self) -> Point2D:
        return [self._DX, self._DY]

    def get_step_minor(self) -> Point2D:
        return [self._dx, self._dy]

    def get_precision(self):
        return self._precision


def geo_graticule() -> Graticule:
    """
    Constructs a geometry generator for creating graticules: a uniform grid of
    meridians and parallels for showing projection distortion. The default
    graticule has meridians and parallels every :math:`10°` between :math:`\\pm
    80°` latitude; for the polar regions, there are meridians every
    :math:`90°`.

    Returns
    -------
    Graticule
        Graticule object
    """
    return Graticule()


def geo_graticule_10() -> GeoJSON:
    """
    A convenience method for directly generating the default :math:`10°` global
    graticule as a GeoJSON MultiLineString geometry object.

    Returns
    -------
    GeoJSON
        GeoJSON object
    """
    return Graticule()()



from datetime import datetime, timedelta

from .interval import TimeInterval


class TimeMinute(TimeInterval):
    """
    Minutes in local time; 60 seconds
    """

    def floor(self, date: datetime) -> datetime:
        """
        Returns a new date representing the latest interval
        boundary date before or equal to date.

        Parameters
        ----------
        date : datetime
            Date

        Returns
        -------
        datetime
            Floored date
        """
        return date.replace(second=0, microsecond=0)

    def offset(self, date: datetime, step: int) -> datetime:
        """
        Returns a new date equal to date plus step intervals.

        Parameters
        ----------
        date : datetime
            Date
        step : int
            Step to add or substitute

        Returns
        -------
        datetime
            Date
        """
        return date + timedelta(minutes=step)

    def count(self, start: datetime, end: datetime) -> int:
        """
        Returns the number of interval boundaries after start
        (exclusive) and before or equal to end (inclusive).

        Parameters
        ----------
        start : datetime
            Start date
        end : datetime
            End date

        Returns
        -------
        int
            Count
        """
        return (end - start).minutes

    def field(self, date: datetime) -> int:
        """
        Returns the field value of the specified date, corresponding to the
        number of boundaries between this date (exclusive) and the latest
        previous parent boundary.

        Parameters
        ----------
        date : datetime
            Floored date

        Returns
        -------
        int
            Field value
        """
        return date.minute


time_minute = TimeMinute()
time_minutes = time_minute.range



from datetime import datetime, timedelta

from .interval import TimeInterval


def time_weekday(i: int):
    class TimeWeekDay(TimeInterval):
        """
        Weeks in local time; typically 7 days
        """

        def floor(self, date: datetime) -> datetime:
            """
            Returns a new date representing the latest interval
            boundary date before or equal to date.

            Parameters
            ----------
            date : datetime
                Date

            Returns
            -------
            datetime
                Floored date
            """
            return date.replace(hour=0, minute=0, second=0, microsecond=0) - timedelta(
                days=(date.weekday() + 7 - (i - 1)) % 7
            )

        def offset(self, date: datetime, step: int) -> datetime:
            """
            Returns a new date equal to date plus step intervals.

            Parameters
            ----------
            date : datetime
                Date
            step : int
                Step to add or substitute

            Returns
            -------
            datetime
                Date
            """
            return date + timedelta(weeks=step)

        def count(self, start: datetime, end: datetime) -> int:
            """
            Returns the number of interval boundaries after start
            (exclusive) and before or equal to end (inclusive).

            Parameters
            ----------
            start : datetime
                Start date
            end : datetime
                End date

            Returns
            -------
            int
                Count
            """
            index = (i + 6) % 7
            a = start.weekday()
            b = end.weekday()
            days = (
                list(range(a, b + 1))
                if a < b
                else list(range(a, 7)) + list(range(b + 1))
            )
            return (end - start).days // 7 + bool(
                index in days and (a != b or index == a)
            )

        def field(self, date: datetime) -> int:
            """
            Returns the field value of the specified date, corresponding to the
            number of boundaries between this date (exclusive) and the latest
            previous parent boundary.

            Parameters
            ----------
            date : datetime
                Floored date

            Returns
            -------
            int
                Field value
            """
            return date.weekday()

    return TimeWeekDay()


time_sunday = time_week = time_weekday(0)
time_monday = time_weekday(1)
time_tuesday = time_weekday(2)
time_wednesday = time_weekday(3)
time_thursday = time_weekday(4)
time_friday = time_weekday(5)
time_saturday = time_weekday(6)

time_sundays = time_weeks = time_sunday.range
time_mondays = time_monday.range
time_tuesdays = time_tuesday.range
time_wednesdays = time_wednesday.range
time_thursdays = time_thursday.range
time_fridays = time_friday.range
time_saturdays = time_saturday.range



from .appearance import order_appearance
from .ascending import order_ascending
from .descending import order_descending
from .inside_out import order_inside_out
from .none import order_none
from .reverse import order_reverse

__all__ = [
    "order_none",
    "order_reverse",
    "order_appearance",
    "order_ascending",
    "order_descending",
    "order_inside_out",
]



# Comments from original code source
# https://github.com/d3/d3-hierarchy/blob/main/src/tree.js
from __future__ import annotations

from collections.abc import Callable

from .hierarchy import Node


class TreeNode(Node):
    def __init__(self, node: Node, i: int):
        super().__init__(None)
        self.node = node
        self.parent = None
        self.children = None
        self.A = None  # default ancestor
        self.a = self  # ancestor
        self.z = 0  # prelim
        self.m = 0  # mod
        self.c = 0  # change
        self.s = 0  # shift
        self.t = None  # thread
        self.i = i  # number


def default_separation(a: TreeNode, b: TreeNode) -> int:
    return 1 if a.parent == b.parent else 2


def next_left(v: TreeNode) -> TreeNode:
    """
    This function is used to traverse the left contour of a subtree (or
    subforest). It returns the successor of v on this contour. This successor
    is either given by the leftmost child of v or by the thread of v. The
    function returns null if and only if v is on the highest level of its
    subtree.
    """
    children = v.children
    return children[0] if children else v.t


def next_right(v: TreeNode) -> TreeNode:
    """
    This function works analogously to next_left.
    """
    children = v.children
    return children[-1] if children else v.t


def move_subtree(wm: TreeNode, wp: TreeNode, shift: float):
    """
    Shifts the current subtree rooted at w+. This is done by increasing
    prelim(w+) and mod(w+) by shift.
    """
    change = shift / (wp.i - wm.i)
    wp.c -= change
    wp.s += shift
    wm.c += change
    wp.z += shift
    wp.m += shift


def execute_shifts(v: TreeNode):
    """
    All other shifts, applied to the smaller subtrees between w- and w+, are
    performed by this function. To prepare the shifts, we have to adjust
    change(w+), shift(w+), and change(w-).
    """
    shift = 0
    change = 0
    for w in reversed(v.children):
        w.z += shift
        w.m += shift
        change += w.c
        shift += w.s + change


def next_ancestor(vim: TreeNode, v: TreeNode, ancestor: TreeNode) -> TreeNode:
    """
    If vi-'s ancestor is a sibling of v, returns vi-'s ancestor. Otherwise,
    returns the specified (default) ancestor.
    """
    return vim.a if vim.a.parent == v.parent else ancestor


def tree_root(root: Node) -> TreeNode:
    tree = TreeNode(root, 0)
    nodes = [tree]

    while nodes:
        node = nodes.pop()
        if children := node.node.children:
            node.children = [None] * len(children)
            for i, c in reversed(list(enumerate(children))):
                child = node.children[i] = TreeNode(c, i)
                child.parent = node
                nodes.append(child)

    parent = tree.parent = TreeNode(None, 0)
    parent.children = [tree]
    return tree


# Node-link tree diagram using the Reingold-Tilford "tidy" algorithm
class Tree:
    """
    Tree layout
    """

    def __init__(self):
        self._separation = default_separation
        self._dx = 1
        self._dy = 1
        self._node_size = None

    def __call__(self, root: Node) -> Node:
        """
        Lays out the specified root hierarchy, assigning the following
        properties on root and its descendants:

        * :code:`node.x` - the x-coordinate of the node
        * :code:`node.y` - the y coordinate of the node

        The coordinates :code:`x` and :code:`y` represent an arbitrary
        coordinate system; for example, you can treat :code:`x` as an angle and
        :code:`y` as a radius to produce a radial layout. You may want to call
        :code:`root.sort` before passing the hierarchy to the tree layout.

        Parameters
        ----------
        root : Node
            Root node

        Returns
        -------
        Node
            Node organized as tree
        """
        t = tree_root(root)

        # Compute the layout using Buchheim et al.'s algorithm.
        t.each_after(self._first_walk)
        t.parent.m = -t.z
        t.each_before(self._second_walk)

        # If a fixed node size is specified, scale x and y.
        if self._node_size:
            root.each_before(self._size_node)
        else:
            # If a fixed tree size is specified, scale x and y based on the
            # extent. Compute the left-most, right-most, and depth-most nodes
            # for extents.
            left = root
            right = root
            bottom = root

            def walk(node: Node):
                nonlocal left, right, bottom
                if node.x < left.x:
                    left = node
                if node.x > right.x:
                    right = node
                if node.depth > bottom.depth:
                    bottom = node

            root.each_before(walk)
            s = 1 if left == right else self._separation(left, right) * 0.5
            tx = s - left.x
            kx = self._dx / (right.x + s + tx)
            ky = self._dy / (bottom.depth or 1)

            def walk(node: Node):
                node.x = (node.x + tx) * kx
                node.y = node.depth * ky

            root.each_before(walk)
        return root

    def _first_walk(self, v: TreeNode):
        """
        Computes a preliminary x-coordinate for v. Before that, FIRST WALK is
        applied recursively to the children of v, as well as the function
        APPORION. After spacing out the children by calling EXECUTE SHIFTS,
        the node v is placed to the midpoint of its outermost children.
        """
        children = v.children
        siblings = v.parent.children
        w = siblings[v.i - 1] if v.i else None

        if children:
            execute_shifts(v)
            midpoint = (children[0].z + children[-1].z) * 0.5
            if w:
                v.z = w.z + self._separation(v.node, w.node)
                v.m = v.z - midpoint
            else:
                v.z = midpoint
        elif w:
            v.z = w.z + self._separation(v.node, w.node)
        v.parent.A = self._apportion(v, w, v.parent.A or siblings[0])

    def _second_walk(self, v: TreeNode):
        """
        Computes all real x-coordinates by summing up the modifiers
        recursively.
        """
        v.node.x = v.z + v.parent.m
        v.m += v.parent.m

    def _apportion(self, v: TreeNode, w: TreeNode, ancestor: TreeNode) -> TreeNode:
        """
        The core of the algorithm. Here, a new subtree is combined with the
        previous subtrees. Threads are used to traverse the inside and outside
        contours of the left and right subtree up to the highest common level.
        The vertices used for the traversals are vi+, vi-, vo-, and vo+, where
        the superscript o means outside and i means inside, the subscript -
        means left subtree and + means right subtree. For summing up the
        modifiers along the contour, we use respective variables si+, si-, so-,
        and so+. Whenever two nodes of the inside contours conflict, we compute
        the left one of the greatest uncommon ancestors using the function
        ANCESOR and call MOVE SUBREE to shift the subtree and prepare the
        shifts of smaller subtrees. Finally, we add a new thread (if
        necessary).
        """
        if w:
            vip = v
            vop = v
            vim = w
            vom = vip.parent.children[0]
            sip = vip.m
            sop = vop.m
            sim = vim.m
            som = vom.m

            vim = next_right(vim)
            vip = next_left(vip)
            while vim and vip:
                vom = next_left(vom)
                vop = next_right(vop)
                vop.a = v
                shift = vim.z + sim - vip.z - sip + self._separation(vim.node, vip.node)
                if shift > 0:
                    move_subtree(next_ancestor(vim, v, ancestor), v, shift)
                    sip += shift
                    sop += shift
                sim += vim.m
                sip += vim.m
                som += vim.m
                sop += vim.m
                vim = next_right(vim)
                vip = next_left(vip)
            if vim and not next_right(vop):
                vop.t = vim
                vim.m += sim - sop
            if vip and not next_left(vom):
                vom.t = vip
                vom.m += sip - som
                ancestor = v
        return ancestor

    def _size_node(self, node: Node):
        node.x *= self._dx
        node.y = node.depth * self._dy

    def set_separation(self, separation: Callable[[Node, Node], int]) -> Tree:
        """
        Sets the :code:`separation` accessor to the specified function and
        returns this tree layout.

        Parameters
        ----------
        separation : Callable[[Node, Node], int]
            Separation function which compares two nodes.

        Returns
        -------
        Tree
            Itself
        """
        self._separation = separation
        return self

    def set_size(self, size: tuple[float, float]) -> Tree:
        """
        Sets this tree layout's size to the specified two-element array of
        numbers :code:`[width, height]` and returns this tree layout. A layout
        size of :code:`None` indicates that a node size will be used instead.
        The coordinates :code:`x` and :code:`y` represent an arbitrary
        coordinate system; for example, to produce a radial layout, a size of
        :code:`[360, radius]` corresponds to a breadth of 360° and a depth of
        radius.

        Parameters
        ----------
        size : tuple[float, float]
            Tree layout's size

        Returns
        -------
        Tree
            Itself
        """
        self._node_size = False
        self._dx = size[0]
        self._dy = size[1]

    def set_node_size(self, node_size: tuple[float, float]) -> Tree:
        """
        Sets this tree layout's node size to the specified two-element array of
        numbers :code:`[width, height]` and returns this tree layout. A node
        size of :code:`None` indicates that a layout size will be used instead.
        When a node size is specified, the root node is always positioned at
        :math:`(0, 0)`.

        Parameters
        ----------
        node_size : tuple[float, float]
            Tree layout's node size

        Returns
        -------
        Tree
            Itself
        """
        self._node_size = True
        self._dx = node_size[0]
        self._dy = node_size[1]
        return self

    def get_separation(self) -> Callable[[Node, Node], int]:
        return self._separation

    def get_size(self) -> tuple[float, float] | None:
        if self._node_size:
            return
        return [self._dx, self._dy]

    def get_node_size(self) -> tuple[float, float] | None:
        if self._node_size:
            return [self._dx, self._dy]
        return None


def tree() -> Tree:
    """
    Builds a new tree layout with default settings.

    Returns
    -------
    Tree
        Tree object
    """
    return Tree()



from ..colors import colors
from ..ramp import ramp

scheme = [
    "ef8a62ffffff999999",
    "ca0020f4a582bababa404040",
    "ca0020f4a582ffffffbababa404040",
    "b2182bef8a62fddbc7e0e0e09999994d4d4d",
    "b2182bef8a62fddbc7ffffffe0e0e09999994d4d4d",
    "b2182bd6604df4a582fddbc7e0e0e0bababa8787874d4d4d",
    "b2182bd6604df4a582fddbc7ffffffe0e0e0bababa8787874d4d4d",
    "67001fb2182bd6604df4a582fddbc7e0e0e0bababa8787874d4d4d1a1a1a",
    "67001fb2182bd6604df4a582fddbc7ffffffe0e0e0bababa8787874d4d4d1a1a1a",
]
SCHEME_RDGY = list(map(colors, scheme))
interpolate_rdgy = ramp(SCHEME_RDGY)



from collections.abc import Callable
from inspect import signature
from operator import itemgetter
from typing import Any


class argpass:
    """
    Determines the number of arguments needed by the given function and returns
    a wrapper function which selects the correct number of arguments.

    Parameters
    ----------
    func : Callable[[list[Any]], Any]
        Function to decorate
    """

    def __init__(self, func: Callable[[list[Any]], Any]):
        self._func = func
        self._nargs = (
            1 if isinstance(func, itemgetter) else len(signature(func).parameters)
        )

    def __call__(self, *args: Any) -> Any:
        """
        Calls the function with the correct number of arguments

        Parameters
        ----------
        *args : Any
            Arbitrary arguments

        Returns
        -------
        Any
            Result of the function call
        """
        return self._func(*args[: self._nargs])

    def __eq__(self, other: Any) -> bool:
        """
        Returns :code:`True` if the inner function equals :code:`other`.

        Parameters
        ----------
        other : Any
            Input object

        Returns
        -------
        bool
            :code:`True` if the inner function equals :code:`other`
        """
        if isinstance(other, argpass):
            return self._func == other._func
        elif callable(other):
            return self._func == other
        else:
            return False



from math import isnan
from typing import TypeVar

from ..types import Point2D

Quadtree = TypeVar("Quadtree", bound="Quadtree")


def add(tree: Quadtree, x: float, y: float, d: Point2D) -> Quadtree:
    if isnan(x) or isnan(y):
        return tree

    parent = None
    node = tree._root
    leaf = {"data": d}
    x0 = tree._x0
    y0 = tree._y0
    x1 = tree._x1
    y1 = tree._y1

    if not node:
        tree._root = leaf
        return tree

    while isinstance(node, list):
        xm = (x0 + x1) * 0.5
        right = x >= xm
        if right:
            x0 = xm
        else:
            x1 = xm

        ym = (y0 + y1) * 0.5
        bottom = y >= ym
        if bottom:
            y0 = ym
        else:
            y1 = ym

        parent = node
        i = bottom << 1 | right
        node = node[i]
        if node is None:
            parent[i] = leaf
            return tree

    xp = tree._x(node["data"])
    yp = tree._y(node["data"])
    if x == xp and y == yp:
        leaf["next"] = node
        if parent:
            parent[i] = leaf
        else:
            tree._root = leaf
        return tree

    while True:
        if parent:
            parent[i] = [None] * 4
            parent = parent[i]
        else:
            parent = tree._root = [None] * 4

        xm = (x0 + x1) * 0.5
        right = x >= xm
        if right:
            x0 = xm
        else:
            x1 = xm

        ym = (y0 + y1) * 0.5
        bottom = y >= ym
        if bottom:
            y0 = ym
        else:
            y1 = ym

        i = bottom << 1 | right
        j = (yp >= ym) << 1 | (xp >= xm)
        if i != j:
            break

    parent[j] = node
    parent[i] = leaf
    return tree



from __future__ import annotations

from collections.abc import Callable

from ..color import cubehelix as color_cubehelix
from .color import color, hue


class CubeHelixInterpolator:
    def __init__(self, func):
        self.func = func
        self.gamma = 1

    def __call__(self, start: str, end: str) -> Callable[[float], str]:
        start = color_cubehelix(start)
        end = color_cubehelix(end)
        h = self.func(start.h, end.h)
        s = color(start.s, end.s)
        l = color(start.l, end.l)
        opacity = color(start.opacity, end.opacity)

        def interpolate(t):
            start.h = h(t)
            start.s = s(t)
            start.l = l(t**self.gamma)
            start.opacity = opacity(t)
            return str(start)

        return interpolate

    def set_gamma(self, gamma: float) -> CubeHelixInterpolator:
        """
        Sets gamma and returns itself. See `gamma error in picture scaling
        <http://www.ericbrasseur.org/gamma.html?i=1>`_ for more information.

        Parameters
        ----------
        gamma : float
            Gamma value

        Returns
        -------
        CubeHelixInterpolator
            Itself
        """
        self.gamma = gamma
        return self


interpolate_cubehelix = CubeHelixInterpolator(hue)
interpolate_cubehelix.__doc__ = """
    Returns a Cubehelix color space interpolator between the two colors a
    and b using a configurable gamma. If the gamma is not specified, it
    defaults to 1.0. The colors a and b need not be in Cubehelix; they will
    be converted to Cubehelix using d3.cubehelix. If either color’s hue or
    saturation is NaN, the opposing color’s channel value is used. The shortest
    path between hues is used. The return value of the interpolator is an
    RGB string.

    Parameters
    ----------
    a : str
        String a
    b : str
        String b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_cubehelix("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(238, 0, 209)'
    """

interpolate_cubehelix_long = CubeHelixInterpolator(color)
interpolate_cubehelix_long.__doc__ = """
    Like :code:`interpolate_cubehelix`, but does not use the shortest path between hues.

    Parameters
    ----------
    a : str
        String a
    b : str
        String b

    Returns
    -------
    Callable[[float], str]
        Interpolator function

    Examples
    --------

    >>> interpolator = d3.interpolate_cubehelix_long("red", "blue")
    >>> interpolator(0)
    'rgb(255, 0, 0)'
    >>> interpolator(1)
    'rgb(0, 0, 255)'
    >>> interpolator(0.5)
    'rgb(238, 0, 209)'
    """



from collections.abc import Callable
from datetime import datetime


def interpolate_date(a: datetime, b: datetime) -> Callable[[float], datetime]:
    """
    Returns an interpolator between the two dates a and b.

    Parameters
    ----------
    a : datetime
        Date a
    b : datetime
        Date b

    Returns
    -------
    Callable[[float], datetime]
        Interpolator function

    Examples
    --------

    >>> from datetime import datetime
    >>> interpolator = d3.interpolate_date(datetime(2000, 1, 1), datetime(2000, 1, 2))
    >>> interpolator(0)
    datetime.datetime(2000, 1, 1, 0, 0)
    >>> interpolator(1)
    datetime.datetime(2000, 1, 2, 0, 0)
    >>> interpolator(0.5)
    datetime.datetime(2000, 1, 1, 12, 0)
    """
    a, b = a.timestamp(), b.timestamp()

    def interpolate(t):
        d = datetime.fromtimestamp(a * (1 - t) + b * t)
        return d

    return interpolate



from collections.abc import Callable

from ...types import GeoJSON, Point2D
from ..common import Projection
from ..path.bounds import BoundsStream
from ..stream import geo_stream


def fit(
    projection: Projection, fit_bounds: Callable[[float], None], obj: GeoJSON
) -> Projection:
    clip = (
        projection.get_clip_extent() if hasattr(projection, "get_clip_extent") else None
    )
    projection.scale(150).translate([0, 0])
    if clip is not None:
        projection.set_clip_extent(None)
    stream = BoundsStream()
    geo_stream(obj, projection.stream(stream))
    fit_bounds(stream.result())
    if clip is not None:
        projection.set_clip_extent(clip)
    return projection


def fit_extent(
    projection: Projection, extent: tuple[Point2D, Point2D], obj: GeoJSON
) -> Projection:
    def fit_bounds(b: float):
        w = extent[1][0] - extent[0][0]
        h = extent[1][1] - extent[0][1]
        k = min(w / (b[1][0] - b[0][0]), h / (b[1][1] - b[0][1]))
        x = extent[0][0] + (w - k * (b[1][0] + b[0][0])) * 0.5
        y = extent[0][1] + (h - k * (b[1][1] + b[0][1])) * 0.5
        projection.scale(150 * k).translate([x, y])

    return fit(projection, fit_bounds, obj)


def fit_size(projection: Projection, size: Point2D, obj: GeoJSON) -> Projection:
    return fit_extent(projection, [[0, 0], size], obj)


def fit_width(projection: Projection, width: float, obj: GeoJSON) -> Projection:
    def fit_bounds(b: float):
        w = width
        k = w / (b[1][0] - b[0][0])
        x = (w - k * (b[1][0] + b[0][0])) / 2
        y = -k * b[0][1]
        projection.scale(150 * k).translate([x, y])

    return fit(projection, fit_bounds, obj)


def fit_height(projection: Projection, height: float, obj: GeoJSON) -> Projection:
    def fit_bounds(b: float):
        h = height
        k = h / (b[1][1] - b[0][1])
        x = -k * b[0][0]
        y = (h - k * (b[1][1] + b[0][1])) / 2
        projection.scale(150 * k).translate([x, y])

    return fit(projection, fit_bounds, obj)



from ..hierarchy import Node
from .dice import dice
from .slice import slice


def slice_dice(parent: Node, x0: float, y0: float, x1: float, y1: float):
    """
    If the specified :code:`parent` has odd depth, delegates to
    :func:`d3.treemap_slice <detroit.treemap_slice>`; otherwise delegates to
    :func:`d3.treemap_dice <detroit.treemap_dice>`.

    Parameters
    ----------
    parent : Node
        Parent node
    x0 : float
        X-coordinate rectangular edge
    y0 : float
        Y-coordinate rectangular edge
    x1 : float
        X-coordinate rectangular edge
    y1 : float
        Y-coordinate rectangular edge
    """
    if parent.depth & 1:
        slice(parent, x0, y0, x1, y1)
    else:
        dice(parent, x0, y0, x1, y1)



from .time_format import (
    iso_format,
    iso_parse,
    time_format,
    time_format_locale,
    time_parse,
)

__all__ = [
    "iso_format",
    "iso_parse",
    "time_format",
    "time_format_locale",
    "time_parse",
]
